"""Database schema conversion steps."""
# MySQL statements to change database structure.
# Every conversion step between schema versions is represented by a
# dictionary, with a key for all statements and a key for all data loss info.

# Key names.
statements = "statements"
step_summary_dict = "step_summary_dict"

# lost_table = When a table is removed and its data is not stored elsewhere.
lost_table = "LOST_TABLE"

# lost_column = When a column is removed and its data is not stored elsewhere.
lost_column = "LOST_COLUMN"

# inaccurate_column = When data is inserted into a column to fulfill
# data structure requirements, but may not be true data.
inaccurate_column = "INACCURATE_COLUMN"

# lost_data = When subset of data in a column is lost, such as due
# to change in datatype that truncates string lengths.
lost_data = "LOST_DATA"

CONVERSION_STEPS = {

    # Upgrade steps
    "upgrade_0_to_1": {
        statements: [
            """CREATE TABLE `version` (
                  `version` int(11) unsigned NOT NULL,
                  PRIMARY KEY (`version`)
                ) ENGINE=InnoDB DEFAULT CHARSET=latin1;""",
            """INSERT INTO `version` (`version`) VALUES ('0');""",
            """ALTER TABLE `gene` DROP FOREIGN KEY `gene_ibfk_1`;""",
            """ALTER TABLE `gene`
                  ADD CONSTRAINT `gene_ibfk_2`
                  FOREIGN KEY (`PhageID`)
                  REFERENCES `phage` (`PhageID`)
                  ON DELETE CASCADE
                  ON UPDATE CASCADE;""",
            """ALTER TABLE `gene_domain` DROP FOREIGN KEY `gene_domain_ibfk_1`;""",
            """ALTER TABLE `gene_domain`
                  ADD CONSTRAINT `gene_domain_ibfk_1`
                  FOREIGN KEY (`GeneID`)
                  REFERENCES `gene` (`GeneID`)
                  ON DELETE CASCADE
                  ON UPDATE CASCADE;""",
            """ALTER TABLE `pham` DROP FOREIGN KEY `pham_ibfk_1`;""",
            """ALTER TABLE `pham`
                  ADD CONSTRAINT `pham_ibfk_1`
                  FOREIGN KEY (`GeneID`)
                  REFERENCES `gene` (`GeneID`)
                  ON DELETE CASCADE
                  ON UPDATE CASCADE;""",
            """ALTER TABLE `scores_summary` DROP FOREIGN KEY `scores_summary_ibfk_1`;""",
            """ALTER TABLE `scores_summary` DROP FOREIGN KEY `scores_summary_ibfk_2`;""",
            """ALTER TABLE `scores_summary`
                  ADD CONSTRAINT `scores_summary_ibfk_1`
                  FOREIGN KEY (`query`)
                  REFERENCES `gene` (`GeneID`)
                  ON DELETE CASCADE
                  ON UPDATE CASCADE,
                  ADD CONSTRAINT `scores_summary_ibfk_2`
                  FOREIGN KEY (`subject`)
                  REFERENCES `gene` (`GeneID`)
                  ON DELETE CASCADE
                  ON UPDATE CASCADE;""",
            """ALTER TABLE `gene` ADD COLUMN `cdd_status` TINYINT(1) NOT NULL DEFAULT '0' AFTER `blast_status`;"""
            ],
        step_summary_dict: {
            inaccurate_column: [
                "version.version",
                "gene.cdd_status"
                ]
            }
        },

    "upgrade_1_to_2": {
        statements: [
            """ALTER TABLE `phage` ADD COLUMN `AnnotationAuthor` tinyint(1) NOT NULL DEFAULT '0' AFTER `status`;""",
            """ALTER TABLE `phage` ADD COLUMN `Program` varchar(10) DEFAULT NULL AFTER `status`;""",
            """ALTER TABLE `phage` ADD COLUMN `AnnotationQC` tinyint(1) NOT NULL DEFAULT '0' AFTER `status`;""",
            """ALTER TABLE `phage` ADD COLUMN `RetrieveRecord` tinyint(1) NOT NULL DEFAULT '0' AFTER `status`;"""
            ],
        step_summary_dict: {
            inaccurate_column: [
                "phage.AnnotationAuthor",
                "phage.Program",
                "phage.AnnotationQC",
                "phage.RetrieveRecord"
                ]
            }
        },

    "upgrade_2_to_3": {
        statements: [
            """ALTER TABLE `version` ADD COLUMN `schema_version` int(11) unsigned NOT NULL AFTER `version`;""",
            """ALTER TABLE `phage` ADD COLUMN `Subcluster2` varchar(5) DEFAULT NULL AFTER `AnnotationAuthor`;""",
            """ALTER TABLE `phage` ADD COLUMN `Cluster2` varchar(5) DEFAULT NULL AFTER `AnnotationAuthor`;""",
            """ALTER TABLE `phage` ADD COLUMN `TempCluster2` varchar(5) DEFAULT NULL;""",
            """UPDATE `phage` SET `Subcluster2` = `Cluster` WHERE `Cluster` REGEXP '[0-9]$';""",
            """DROP PROCEDURE IF EXISTS split_subcluster;""",
            # Note: below is the syntax if executing directly in mysql:
            # """DELIMITER //
            #     CREATE PROCEDURE split_subcluster()
            #     BEGIN
            #         DECLARE Counter INT DEFAULT 8;
            #         WHILE Counter > -1 DO
            #             UPDATE `phage` SET `TempCluster2` = SUBSTRING(`Subcluster2`, 1, LENGTH(`Subcluster2`)-Counter);
            #             UPDATE `phage` SET `Cluster2` = `TempCluster2` WHERE `TempCluster2` NOT REGEXP '[0-9]$' and `TempCluster2` != '';
            #             SET Counter = Counter - 1;
            #         END WHILE;
            #     END //
            #     DELIMITER;""",
            """CREATE PROCEDURE split_subcluster()
                BEGIN
                    DECLARE Counter INT DEFAULT 8;
                    WHILE Counter > -1 DO
                        UPDATE `phage` SET `TempCluster2` = SUBSTRING(`Subcluster2`, 1, LENGTH(`Subcluster2`)-Counter);
                        UPDATE `phage` SET `Cluster2` = `TempCluster2` WHERE `TempCluster2` NOT REGEXP '[0-9]$' and `TempCluster2` != '';
                        SET Counter = Counter - 1;
                    END WHILE;
                END;""",
            """CALL split_subcluster;""",
            """DROP PROCEDURE split_subcluster;""",
            """UPDATE `phage` SET `Cluster2` = `Cluster` WHERE `Cluster` NOT REGEXP '[0-9]$';""",
            """ALTER TABLE `phage` DROP COLUMN `TempCluster2`;""",
            """ALTER TABLE `phage` DROP COLUMN `Program`;""",
            """ALTER TABLE `gene` ADD COLUMN `LocusTag` varchar(50) DEFAULT NULL AFTER `cdd_status`;""",
            """UPDATE `version` SET `schema_version` = 3;"""
            ],
        step_summary_dict: {
            lost_column: ["phage.Program"],
            inaccurate_column: ["gene.LocusTag"]
            }
        },

    "upgrade_3_to_4": {
        statements: [
            """CREATE TABLE `tmrna` (
                  `PhageID` varchar(25) NOT NULL,
                  `TmrnaID` varchar(35) NOT NULL,
                  `LocusTag` varchar(35) DEFAULT NULL,
                  `Start` mediumint(9) NOT NULL,
                  `Stop` mediumint(9) NOT NULL,
                  `Orientation` enum('F','R') NOT NULL,
                  `Note` blob,
                  `PeptideTag` varchar(50) DEFAULT NULL,
                  PRIMARY KEY (`TmrnaID`)
                ) ENGINE=InnoDB DEFAULT CHARSET=latin1;""",
            """CREATE TABLE `trna` (
                  `PhageID` varchar(25) NOT NULL,
                  `TrnaID` varchar(35) NOT NULL,
                  `LocusTag` varchar(35) DEFAULT NULL,
                  `Start` mediumint(9) NOT NULL,
                  `Stop` mediumint(9) NOT NULL,
                  `Length` mediumint(9) NOT NULL,
                  `Orientation` enum('F','R') NOT NULL,
                  `Sequence` varchar(100) NOT NULL,
                  `Product` blob,
                  `Note` blob,
                  `AminoAcid` enum('Ala','Arg','Asn','Asp','Cys','Gln','Glu','Gly','His','Ile','Leu','Lys','Met','Phe','Pro','Ser','Thr','Trp','Tyr','Val','Undet','OTHER') NOT NULL,
                  `Anticodon` varchar(4) NOT NULL,
                  `InfernalScore` decimal(4,2) DEFAULT NULL,
                  PRIMARY KEY (`TrnaID`),
                  KEY `PhageID` (`PhageID`),
                  FOREIGN KEY (`PhageID`) REFERENCES `phage` (`PhageID`)
                ) ENGINE=InnoDB DEFAULT CHARSET=latin1;""",
            """CREATE TABLE `trna_structures` (
                  `Sequence` varchar(100) NOT NULL,
                  `Structure` varchar(300) NOT NULL,
                  PRIMARY KEY (`Sequence`)
                ) ENGINE=InnoDB DEFAULT CHARSET=latin1;""",
            """ALTER TABLE `gene` MODIFY `translation` VARCHAR(5000) DEFAULT NULL;""",
            """UPDATE `version` SET `schema_version` = 4;"""
            ],
        step_summary_dict: {}
        },

    "upgrade_4_to_5": {
        statements: [
            """DROP TABLE `node`;""",
            """DROP TABLE `host_range`;""",
            """DROP TABLE `host`;""",
            """DROP TABLE `pham_history`;""",
            """DROP TABLE `pham_old`;""",
            """DROP TABLE `scores_summary`;""",
            """ALTER TABLE `phage` DROP COLUMN `Prophage`;""",
            """ALTER TABLE `phage` DROP COLUMN `Isolated`;""",
            """ALTER TABLE `phage` DROP COLUMN `ProphageOffset`;""",
            """ALTER TABLE `phage` DROP COLUMN `DateLastSearched`;""",
            """ALTER TABLE `phage` DROP COLUMN `AnnotationQC`;""",
            """ALTER TABLE `gene` DROP COLUMN `StartCodon`;""",
            """ALTER TABLE `gene` DROP COLUMN `StopCodon`;""",
            """ALTER TABLE `gene` DROP COLUMN `GC1`;""",
            """ALTER TABLE `gene` DROP COLUMN `GC2`;""",
            """ALTER TABLE `gene` DROP COLUMN `GC3`;""",
            """ALTER TABLE `gene` DROP COLUMN `GC`;""",
            """ALTER TABLE `gene` DROP COLUMN `LeftNeighbor`;""",
            """ALTER TABLE `gene` DROP COLUMN `RightNeighbor`;""",
            """ALTER TABLE `gene` DROP COLUMN `clustalw_status`;""",
            """ALTER TABLE `gene` DROP COLUMN `blast_status`;""",
            """ALTER TABLE `gene` DROP COLUMN `TypeID`;""",
            """ALTER TABLE `pham` DROP COLUMN `orderAdded`;""",
            """ALTER TABLE `phage` MODIFY `status` varchar(7);""",
            """UPDATE `phage` SET `status` = 'unknown' WHERE `status` = 'gbk';""",
            """ALTER TABLE `phage` MODIFY `status` enum('unknown','draft','final');""",
            """UPDATE `version` SET `schema_version` = 5;"""
            ],
        step_summary_dict: {
            lost_table: [
                "node",
                "host_range",
                "host",
                "pham_history",
                "pham_old",
                "scores_summary"
                ],
            lost_column: [
                "phage.Prophage",
                "phage.Isolated",
                "phage.ProphageOffset",
                "phage.DateLastSearched",
                "phage.AnnotationQC",
                "gene.StartCodon",
                "gene.StopCodon",
                "gene.GC1",
                "gene.GC2",
                "gene.GC3",
                "gene.GC",
                "gene.LeftNeighbor",
                "gene.RightNeighbor",
                "gene.clustalw_status",
                "gene.blast_status",
                "gene.TypeID",
                "pham.orderAdded"
                ]
            }
        },

    "upgrade_5_to_6": {
        statements: [
            """ALTER TABLE `domain` CHANGE `id` `ID` int(10) unsigned NOT NULL AUTO_INCREMENT;""",
            """ALTER TABLE `domain` CHANGE `description` `Description` blob;""",
            """ALTER TABLE `gene_domain` CHANGE `id` `ID` int(10) unsigned NOT NULL AUTO_INCREMENT;""",
            """ALTER TABLE `gene_domain` CHANGE `expect` `Expect` double unsigned NOT NULL;""",
            """ALTER TABLE `phage` CHANGE `status` `Status` enum('unknown','draft','final') DEFAULT NULL;""",
            """ALTER TABLE `pham` CHANGE `name` `Name` int(10) unsigned DEFAULT NULL;""",
            """ALTER TABLE `pham_color` CHANGE `id` `ID` int(10) unsigned NOT NULL AUTO_INCREMENT;""",
            """ALTER TABLE `pham_color` CHANGE `name` `Name` int(10) unsigned NOT NULL;""",
            """ALTER TABLE `pham_color` CHANGE `color` `Color` char(7) NOT NULL;""",
            """ALTER TABLE `version` CHANGE `version` `Version` int(11) unsigned NOT NULL;""",
            """ALTER TABLE `gene` CHANGE `translation` `Translation` varchar(5000) DEFAULT NULL;""",
            """ALTER TABLE `gene` CHANGE `id` `ID` int(10) unsigned NOT NULL AUTO_INCREMENT;""",
            """ALTER TABLE `gene` CHANGE `cdd_status` `DomainStatus` tinyint(1) NOT NULL DEFAULT '0';""",
            """ALTER TABLE `version` CHANGE `schema_version` `SchemaVersion` int(11) unsigned NOT NULL;""",
            """ALTER TABLE `domain` CHANGE `hit_id` `HitID` varchar(25) NOT NULL;""",
            """ALTER TABLE `gene_domain` CHANGE `hit_id` `HitID` varchar(25) NOT NULL;""",
            """ALTER TABLE `gene_domain` CHANGE `query_start` `QueryStart` int(10) unsigned NOT NULL;""",
            """ALTER TABLE `gene_domain` CHANGE `query_end` `QueryEnd` int(10) unsigned NOT NULL;""",
            """UPDATE `version` SET `SchemaVersion` = 6;"""
            ],
        step_summary_dict: {}
        },

    "upgrade_6_to_7": {
        statements: [
            """ALTER TABLE `phage` CHANGE `HostStrain` `HostGenus` varchar(50) DEFAULT NULL;""",
            """ALTER TABLE `phage` DROP COLUMN `Cluster`;""",
            """ALTER TABLE `phage` CHANGE `Cluster2` `Cluster` varchar(5) DEFAULT NULL;""",
            """ALTER TABLE `phage` CHANGE `Subcluster2` `Subcluster` varchar(5) DEFAULT NULL;""",
            """ALTER TABLE `phage` CHANGE `SequenceLength` `Length` mediumint(9) NOT NULL;""",
            """ALTER TABLE `pham_color` CHANGE `Name` `PhamID` int(10) unsigned NOT NULL;""",
            """ALTER TABLE `pham_color` DROP `ID`;""",
            """ALTER TABLE `pham_color` ADD PRIMARY KEY (`PhamID`);""",
            """ALTER TABLE `gene` DROP `ID`;""",
            """ALTER TABLE `gene` ADD COLUMN `Parts` tinyint(1) DEFAULT NULL;""",
            """ALTER TABLE `gene` ADD COLUMN `PhamID` int(10) unsigned DEFAULT NULL;""",
            """UPDATE `gene`
                LEFT JOIN `pham` ON `gene`.`GeneID` = `pham`.`GeneID`
                SET `gene`.`PhamID` = `pham`.`Name`;""",
            """ALTER TABLE `gene`
                ADD FOREIGN KEY (`PhamID`)
                REFERENCES `pham_color` (`PhamID`);""",
            """DROP TABLE `pham`;""",
            """RENAME TABLE `pham_color` TO `pham`;""",
            """UPDATE `version` SET `SchemaVersion` = 7;"""
            ],
        step_summary_dict: {
            inaccurate_column: ["gene.Parts"]
            }
        },

    "upgrade_7_to_8": {
        statements: [
            """ALTER TABLE `gene` DROP FOREIGN KEY `gene_ibfk_3`;""",
            """ALTER TABLE `gene`
                  ADD FOREIGN KEY (`PhamID`)
                  REFERENCES `pham` (`PhamID`)
                  ON DELETE SET NULL;""",
            """UPDATE `version` SET `SchemaVersion` = 8;"""
            ],
        step_summary_dict: {}
        },

    "upgrade_8_to_9": {
        statements: [
            """ALTER TABLE `trna` ADD COLUMN `Structure` varchar(300) NOT NULL;""",
            """UPDATE `trna`
              LEFT JOIN `trna_structures` ON `trna`.`Sequence` = `trna_structures`.`Sequence`
              SET `trna`.`Structure` = `trna_structures`.`Structure`;""",
            """ALTER TABLE `trna` MODIFY COLUMN `Structure` blob DEFAULT NULL;""",
            """DROP TABLE `trna_structures`;""",
            """UPDATE `trna` SET `AminoAcid` = 'OTHER' WHERE `AminoAcid` = 'Undet';""",
            """ALTER TABLE `trna` MODIFY COLUMN `AminoAcid` enum('Ala','Arg','Asn','Asp','Cys','fMet','Gln','Glu','Gly','His','Ile','Ile2','Leu','Lys','Met','Phe','Pro','Pyl','SeC','Ser','Thr','Trp','Tyr','Val','Stop','OTHER') NOT NULL;""",
            """ALTER TABLE `trna` DROP COLUMN `Sequence`;""",
            """ALTER TABLE `trna` DROP COLUMN `Product`;""",
            """ALTER TABLE `trna` DROP COLUMN `InfernalScore`;""",
            """ALTER TABLE `trna` ADD COLUMN `Source` enum('aragorn', 'trnascan', 'both') DEFAULT NULL;""",
            """ALTER TABLE `trna` MODIFY COLUMN `PhageID` varchar(25) NOT NULL AFTER `TrnaID`;""",
            """ALTER TABLE `trna` MODIFY COLUMN `LocusTag` varchar(35) DEFAULT NULL AFTER `Note`;""",
            """ALTER TABLE `trna` CHANGE `TrnaID` `GeneID` varchar(35) NOT NULL;""",
            """ALTER TABLE `trna` ADD COLUMN `Name` varchar(50) NOT NULL AFTER `Length`;""",
            """ALTER TABLE `trna` DROP FOREIGN KEY `trna_ibfk_1`;""",
            """ALTER TABLE `trna`
                  ADD CONSTRAINT `trna_ibfk_1`
                  FOREIGN KEY (`PhageID`)
                  REFERENCES `phage` (`PhageID`)
                  ON DELETE CASCADE
                  ON UPDATE CASCADE;""",
            """ALTER TABLE `tmrna` ADD KEY (`PhageID`);""",
            """ALTER TABLE `tmrna`
                  ADD CONSTRAINT `tmrna_ibfk_1`
                  FOREIGN KEY (`PhageID`)
                  REFERENCES `phage` (`PhageID`)
                  ON DELETE CASCADE
                  ON UPDATE CASCADE;""",
            """ALTER TABLE `tmrna` MODIFY COLUMN `PhageID` varchar(25) NOT NULL AFTER `TmrnaID`;""",
            """ALTER TABLE `tmrna` MODIFY COLUMN `LocusTag` varchar(35) DEFAULT NULL AFTER `Note`;""",
            """ALTER TABLE `tmrna` CHANGE `TmrnaID` `GeneID` varchar(35) NOT NULL;""",
            """ALTER TABLE `tmrna` ADD COLUMN `Length` mediumint(9) NOT NULL AFTER `Stop`;""",
            """ALTER TABLE `tmrna` ADD COLUMN `Name` varchar(50) NOT NULL AFTER `Length`;""",
            """UPDATE `version` SET `SchemaVersion` = 9;"""
            ],
        step_summary_dict: {
            lost_column: [
                "trna.Sequence",
                "trna.Product",
                "trna.InfernalScore"
                ],
            inaccurate_column: [
                "trna.Source",
                "trna.Name",
                "tmrna.Name",
                "tmrna.Length"
                ]
            }
        },

    "upgrade_9_to_10": {
        statements: [
            """ALTER TABLE `gene` MODIFY COLUMN `Translation` blob DEFAULT NULL;""",
            """UPDATE `version` SET `SchemaVersion` = 10;"""
            ],
        step_summary_dict: {}
        },

    # Downgrade steps
    "downgrade_1_to_0": {
        statements: [
            """ALTER TABLE `gene` DROP COLUMN `cdd_status`;""",
            """ALTER TABLE `scores_summary` DROP FOREIGN KEY `scores_summary_ibfk_1`;""",
            """ALTER TABLE `scores_summary` DROP FOREIGN KEY `scores_summary_ibfk_2`;""",
            """ALTER TABLE `scores_summary`
                  ADD CONSTRAINT `scores_summary_ibfk_1`
                  FOREIGN KEY (`query`)
                  REFERENCES `gene` (`GeneID`),
                  ADD CONSTRAINT `scores_summary_ibfk_2`
                  FOREIGN KEY (`subject`)
                  REFERENCES `gene` (`GeneID`);""",
            """ALTER TABLE `pham` DROP FOREIGN KEY `pham_ibfk_1`;""",
            """ALTER TABLE `pham`
                  ADD CONSTRAINT `pham_ibfk_1`
                  FOREIGN KEY (`GeneID`)
                  REFERENCES `gene` (`GeneID`);""",
            """ALTER TABLE `gene_domain` DROP FOREIGN KEY `gene_domain_ibfk_1`;""",
            """ALTER TABLE `gene_domain`
                  ADD CONSTRAINT `gene_domain_ibfk_1`
                  FOREIGN KEY (`GeneID`)
                  REFERENCES `gene` (`GeneID`);""",
            """ALTER TABLE `gene` DROP FOREIGN KEY `gene_ibfk_2`;""",
            """ALTER TABLE `gene`
                  ADD CONSTRAINT `gene_ibfk_1`
                  FOREIGN KEY (`PhageID`)
                  REFERENCES `phage` (`PhageID`);""",
            """DROP TABLE `version`;"""
            ],
        step_summary_dict: {
            lost_table: ["version"],
            lost_column: ["gene.cdd_status"]
            }
        },

    "downgrade_2_to_1": {
        statements: [
            """ALTER TABLE `phage` DROP COLUMN `RetrieveRecord`;""",
            """ALTER TABLE `phage` DROP COLUMN `AnnotationQC`;""",
            """ALTER TABLE `phage` DROP COLUMN `Program`;""",
            """ALTER TABLE `phage` DROP COLUMN `AnnotationAuthor`;"""
            ],
        step_summary_dict: {
            lost_column: [
                "phage.RetrieveRecord",
                "phage.AnnotationQC",
                "phage.Program",
                "phage.AnnotationAuthor"
                ]
            }
        },

    "downgrade_3_to_2": {
        statements: [
            """ALTER TABLE `phage` ADD COLUMN `Program` varchar(10) DEFAULT NULL AFTER `AnnotationQC`;""",
            """ALTER TABLE `phage` DROP COLUMN `Cluster2`;""",
            """ALTER TABLE `phage` DROP COLUMN `Subcluster2`;""",
            """ALTER TABLE `gene` DROP COLUMN `LocusTag`;""",
            """ALTER TABLE `version` DROP COLUMN `schema_version`;""",
            ],
        step_summary_dict: {
            lost_column: [
                "gene.LocusTag",
                "version.schema_version"
                ],
            inaccurate_column: ["phage.Program"]
            }
        },

    "downgrade_4_to_3": {
        statements: [
            """UPDATE `gene` SET `translation` = NULL WHERE LENGTH(`translation`) > 4000;""",
            """ALTER TABLE `gene` MODIFY `translation` VARCHAR(4000) DEFAULT NULL;""",
            """DROP TABLE `tmrna`;""",
            """DROP TABLE `trna`;""",
            """DROP TABLE `trna_structures`;""",
            """UPDATE `version` SET `schema_version` = 3;"""
            ],
        step_summary_dict: {
            lost_table: [
                "tmrna",
                "trna",
                "trna_structures"
                ],
            lost_data: ["gene.translation"]
            }
        },

    "downgrade_5_to_4": {
        statements: [
            """ALTER TABLE `phage` MODIFY `status` varchar(7) DEFAULT NULL;""",
            """UPDATE `phage` SET `status` = 'gbk' WHERE `status` = 'unknown';""",
            """ALTER TABLE `phage` MODIFY `status` varchar(5) DEFAULT NULL;""",
            """ALTER TABLE `pham` ADD COLUMN `orderAdded` int(5) unsigned NOT NULL AFTER `Name`;""",
            """ALTER TABLE `pham` ADD KEY (`orderAdded`);""",
            """ALTER TABLE `pham` MODIFY `orderAdded` int(5) unsigned NOT NULL AUTO_INCREMENT;""",
            """ALTER TABLE `gene` ADD COLUMN `TypeID` varchar(10) DEFAULT NULL AFTER `Name`;""",
            """UPDATE `gene` SET `TypeID` = 'CDS';""",
            """ALTER TABLE `gene` ADD COLUMN `blast_status` enum('avail','pending','stale','done') NOT NULL DEFAULT 'avail' AFTER `id`;""",
            """ALTER TABLE `gene` ADD COLUMN `clustalw_status` enum('avail','pending','stale','done') NOT NULL DEFAULT 'avail' AFTER `id`;""",
            """ALTER TABLE `gene` ADD COLUMN `RightNeighbor` varchar(25) DEFAULT NULL AFTER `Orientation`;""",
            """ALTER TABLE `gene` ADD COLUMN `LeftNeighbor` varchar(25) DEFAULT NULL AFTER `Orientation`;""",
            """ALTER TABLE `gene` ADD COLUMN `GC` float DEFAULT NULL AFTER `Orientation`;""",
            """ALTER TABLE `gene` ADD COLUMN `GC3` float DEFAULT NULL AFTER `Orientation`;""",
            """ALTER TABLE `gene` ADD COLUMN `GC2` float DEFAULT NULL AFTER  `Orientation`;""",
            """ALTER TABLE `gene` ADD COLUMN `GC1` float DEFAULT NULL AFTER `Orientation`;""",
            """ALTER TABLE `gene` ADD COLUMN `StopCodon` enum('TAA','TAG','TGA') DEFAULT NULL AFTER `translation`;""",
            """ALTER TABLE `gene` ADD COLUMN `StartCodon` enum('ATG','GTG','TTG') DEFAULT NULL AFTER `translation`;""",
            """ALTER TABLE `phage` ADD COLUMN `AnnotationQC` tinyint(1) NOT NULL DEFAULT '0' AFTER `RetrieveRecord`;""",
            """ALTER TABLE `phage` ADD COLUMN `DateLastSearched` datetime DEFAULT NULL AFTER `DateLastModified`;""",
            """ALTER TABLE `phage` ADD COLUMN `ProphageOffset` int(11) DEFAULT NULL AFTER `SequenceLength`;""",
            """ALTER TABLE `phage` ADD COLUMN `Prophage` enum('yes','no') DEFAULT NULL AFTER `SequenceLength`;""",
            """ALTER TABLE `phage` ADD COLUMN `Isolated` varchar(100) DEFAULT NULL AFTER `Name`;""",
            """CREATE TABLE `scores_summary` (
                  `id` bigint(20) NOT NULL AUTO_INCREMENT,
                  `query` varchar(35) DEFAULT NULL,
                  `subject` varchar(35) DEFAULT NULL,
                  `blast_score` double unsigned DEFAULT NULL,
                  `clustalw_score` decimal(5,4) unsigned DEFAULT NULL,
                  `blast_bit_score` double unsigned DEFAULT NULL,
                  PRIMARY KEY (`id`),
                  KEY (`query`),
                  KEY (`subject`),
                  FOREIGN KEY (`query`) REFERENCES `gene` (`GeneID`) ON DELETE CASCADE ON UPDATE CASCADE,
                  FOREIGN KEY (`subject`) REFERENCES `gene` (`GeneID`) ON DELETE CASCADE ON UPDATE CASCADE
                ) ENGINE=InnoDB DEFAULT CHARSET=latin1;""",
            """CREATE TABLE `pham_old` (
                  `GeneID` varchar(35) DEFAULT NULL,
                  `name` int(10) unsigned DEFAULT NULL,
                  `orderAdded` int(5) unsigned NOT NULL AUTO_INCREMENT,
                  PRIMARY KEY (`orderAdded`),
                  KEY (`orderAdded`),
                  KEY (`GeneID`),
                  FOREIGN KEY (`GeneID`) REFERENCES `gene` (`GeneID`) ON UPDATE CASCADE
                ) ENGINE=InnoDB DEFAULT CHARSET=latin1;""",
            """CREATE TABLE `pham_history` (
                  `name` int(10) unsigned NOT NULL,
                  `parent` int(10) unsigned NOT NULL,
                  `action` enum('join','split') NOT NULL,
                  `datetime` timestamp NOT NULL DEFAULT CURRENT_TIMESTAMP,
                  PRIMARY KEY (`name`,`parent`,`action`),
                  KEY (`parent`)
                ) ENGINE=InnoDB DEFAULT CHARSET=latin1;""",
            """CREATE TABLE `host` (
                  `ID` int(11) NOT NULL AUTO_INCREMENT,
                  `Name` varchar(50) NOT NULL,
                  `Accession` varchar(15) DEFAULT NULL,
                  PRIMARY KEY (`ID`)
                ) ENGINE=InnoDB DEFAULT CHARSET=latin1;""",
            """CREATE TABLE `host_range` (
                  `PhageID` varchar(25) NOT NULL,
                  `host_id` int(11) NOT NULL,
                  `plating_efficiency` double unsigned NOT NULL,
                  `below_detection` tinyint(1) NOT NULL,
                  KEY (`PhageID`),
                  KEY (`host_id`),
                  FOREIGN KEY (`PhageID`) REFERENCES phage (`PhageID`),
                  FOREIGN KEY (`host_id`) REFERENCES host (`ID`)
                ) ENGINE=InnoDB DEFAULT CHARSET=latin1;""",
            """CREATE TABLE `node` (
                  `id` int(11) unsigned NOT NULL AUTO_INCREMENT,
                  `platform` varchar(15) DEFAULT NULL,
                  `hostname` varchar(50) DEFAULT NULL,
                  PRIMARY KEY (`id`),
                  KEY (`hostname`)
                ) ENGINE=InnoDB DEFAULT CHARSET=latin1;""",
            """UPDATE `version` SET `schema_version` = 4;"""
            ],
        step_summary_dict: {
            inaccurate_column: [
                "pham.orderAdded",
                "gene.blast_status",
                "gene.clustalw_status",
                "gene.RightNeighbor",
                "gene.LeftNeighbor",
                "gene.GC",
                "gene.GC3",
                "gene.GC2",
                "gene.GC1",
                "gene.StopCodon",
                "gene.StartCodon",
                "phage.AnnotationQC",
                "phage.DateLastSearched",
                "phage.ProphageOffset",
                "phage.Prophage",
                "phage.Isolated"
                ]
            }
        },

    "downgrade_6_to_5": {
        statements: [
            """ALTER TABLE `gene_domain` CHANGE `QueryEnd` `query_end` int(10) unsigned NOT NULL;""",
            """ALTER TABLE `gene_domain` CHANGE `QueryStart` `query_start` int(10) unsigned NOT NULL;""",
            """ALTER TABLE `gene_domain` CHANGE `HitID` `hit_id` varchar(25) NOT NULL;""",
            """ALTER TABLE `domain` CHANGE `HitID` `hit_id` varchar(25) NOT NULL;""",
            """ALTER TABLE `version` CHANGE `SchemaVersion` `schema_version` int(11) unsigned NOT NULL;""",
            """ALTER TABLE `gene` CHANGE `DomainStatus` `cdd_status` tinyint(1) NOT NULL DEFAULT '0';""",
            """ALTER TABLE `gene` CHANGE `ID` `id` int(10) unsigned NOT NULL AUTO_INCREMENT;""",
            """ALTER TABLE `gene` CHANGE `Translation` `translation` varchar(5000) DEFAULT NULL;""",
            """ALTER TABLE `version` CHANGE `Version` `version` int(11) unsigned NOT NULL;""",
            """ALTER TABLE `pham_color` CHANGE `Color` `color` char(7) NOT NULL;""",
            """ALTER TABLE `pham_color` CHANGE `Name` `name` int(10) unsigned NOT NULL;""",
            """ALTER TABLE `pham_color` CHANGE `ID` `id` int(10) unsigned NOT NULL AUTO_INCREMENT;""",
            """ALTER TABLE `pham` CHANGE `Name` `name` int(10) unsigned DEFAULT NULL;""",
            """ALTER TABLE `phage` CHANGE `Status` `status` enum('unknown','draft','final') DEFAULT NULL;""",
            """ALTER TABLE `gene_domain` CHANGE `Expect` `expect` double unsigned NOT NULL;""",
            """ALTER TABLE `gene_domain` CHANGE `ID` `id` int(10) unsigned NOT NULL AUTO_INCREMENT;""",
            """ALTER TABLE `domain` CHANGE `Description` `description` blob;""",
            """ALTER TABLE `domain` CHANGE `ID` `id` int(10) unsigned NOT NULL AUTO_INCREMENT;""",
            """UPDATE `version` SET `schema_version` = 5;"""
            ],
        step_summary_dict: {}
        },

    "downgrade_7_to_6": {
        statements: [
            """RENAME TABLE `pham` TO `pham_color`;""",
            """CREATE TABLE `pham` (
                  `GeneID` varchar(35) NOT NULL DEFAULT '',
                  `Name` int(10) unsigned DEFAULT NULL,
                  PRIMARY KEY (`GeneID`),
                  KEY `name_index` (`Name`),
                  CONSTRAINT `pham_ibfk_1` FOREIGN KEY (`GeneID`) REFERENCES `gene` (`GeneID`) ON DELETE CASCADE ON UPDATE CASCADE
                  ) ENGINE=InnoDB DEFAULT CHARSET=latin1;""",
            """ALTER TABLE `gene` DROP FOREIGN KEY `gene_ibfk_3`;""",
            """INSERT INTO `pham` (`GeneID`, `Name`)
                    SELECT `GeneID`,`PhamID` FROM `gene`;""",
            """ALTER TABLE `gene` DROP `PhamID`;""",
            """ALTER TABLE `gene` DROP `Parts`;""",
            """ALTER TABLE `gene` ADD COLUMN `ID` int(10) unsigned NOT NULL AFTER `Notes`;""",
            """ALTER TABLE `gene` ADD KEY `id` (`ID`);""",
            """ALTER TABLE `gene` MODIFY `ID` int(10) unsigned NOT NULL AUTO_INCREMENT;""",
            """ALTER TABLE `pham_color` DROP PRIMARY KEY;""",
            """ALTER TABLE `pham_color` ADD COLUMN `ID` int(10) unsigned NOT NULL FIRST;""",
            """ALTER TABLE `pham_color` ADD KEY (`ID`);""",
            """ALTER TABLE `pham_color` MODIFY `ID` int(10) unsigned NOT NULL AUTO_INCREMENT;""",
            """ALTER TABLE `pham_color` ADD PRIMARY KEY (`ID`);""",
            """ALTER TABLE `pham_color` DROP KEY `ID`;""",
            """ALTER TABLE `pham_color` CHANGE `PhamID` `Name` int(10) unsigned NOT NULL;""",
            """ALTER TABLE `phage` CHANGE `Length` `SequenceLength` mediumint(9) NOT NULL;""",
            """ALTER TABLE `phage` CHANGE `Subcluster` `Subcluster2` varchar(5) DEFAULT NULL;""",
            """ALTER TABLE `phage` CHANGE `Cluster` `Cluster2` varchar(5) DEFAULT NULL;""",
            """ALTER TABLE `phage` ADD COLUMN `Cluster` varchar(5) DEFAULT NULL AFTER `GC`;""",
            """UPDATE `phage` SET `Cluster` = `Subcluster2`;""",
            """UPDATE `phage` SET `Cluster` = `Cluster2` WHERE `Cluster` is NULL;""",
            """ALTER TABLE `phage` CHANGE `HostGenus` `HostStrain` varchar(50) DEFAULT NULL;""",
            """UPDATE `version` SET `SchemaVersion` = 6;"""
            ],
        step_summary_dict: {
            lost_column: ["gene.Parts"]
            }
        },

    "downgrade_8_to_7": {
        statements: [
            """ALTER TABLE `gene` DROP FOREIGN KEY `gene_ibfk_3`;""",
            """ALTER TABLE `gene`
                  ADD FOREIGN KEY (`PhamID`)
                  REFERENCES `pham` (`PhamID`);""",
            """UPDATE `version` SET `SchemaVersion` = 7;"""
            ],
        step_summary_dict: {}
        },

    "downgrade_9_to_8": {
        statements: [
            """ALTER TABLE `tmrna` DROP COLUMN `Name`;""",
            """ALTER TABLE `tmrna` DROP COLUMN `Length`;""",
            """ALTER TABLE `tmrna` CHANGE `GeneID` `TmrnaID` varchar(35) NOT NULL;""",
            """ALTER TABLE `tmrna` MODIFY COLUMN `LocusTag` varchar(35) DEFAULT NULL AFTER `PhageID`;""",
            """ALTER TABLE `tmrna` MODIFY COLUMN `TmrnaID` varchar(35) NOT NULL AFTER `PhageID`;""",
            """ALTER TABLE `tmrna` DROP FOREIGN KEY `tmrna_ibfk_1`;""",
            """ALTER TABLE `tmrna` DROP KEY `PhageID`;""",
            """ALTER TABLE `trna` DROP FOREIGN KEY `trna_ibfk_1`;""",
            """ALTER TABLE `trna`
              ADD CONSTRAINT `trna_ibfk_1`
              FOREIGN KEY (`PhageID`)
              REFERENCES `phage` (`PhageID`);""",
            """ALTER TABLE `trna` DROP COLUMN `Name`;""",
            """ALTER TABLE `trna` CHANGE `GeneID` `TrnaID` varchar(35) NOT NULL;""",
            """ALTER TABLE `trna` MODIFY COLUMN `LocusTag` varchar(35) DEFAULT NULL AFTER `PhageID`;""",
            """ALTER TABLE `trna` MODIFY COLUMN `TrnaID` varchar(35) NOT NULL AFTER `PhageID`;""",
            """ALTER TABLE `trna` DROP COLUMN `Source`;""",
            """ALTER TABLE `trna` ADD COLUMN `InfernalScore` decimal(4,2) DEFAULT NULL AFTER `Anticodon`;""",
            """ALTER TABLE `trna` ADD COLUMN `Product` blob AFTER `Orientation`;""",
            """ALTER TABLE `trna` ADD COLUMN `Sequence` varchar(100) DEFAULT NULL AFTER `Orientation`;""",
            """UPDATE `trna` SET `Sequence` = '';""",
            """ALTER TABLE `trna` MODIFY COLUMN `Sequence` varchar(100) NOT NULL;""",
            """UPDATE `trna` SET `AminoAcid` = 'OTHER' WHERE `AminoAcid` in ('fMet', 'Ile2', 'Pyl', 'Sec', 'Stop');""",
            """ALTER TABLE `trna` MODIFY COLUMN `AminoAcid` enum('Ala','Arg','Asn','Asp','Cys','Gln','Glu','Gly','His','Ile','Leu','Lys','Met','Phe','Pro','Ser','Thr','Trp','Tyr','Val','Undet','OTHER') NOT NULL;""",
            """CREATE TABLE `trna_structures` (
              `Sequence` varchar(100) NOT NULL,
              `Structure` varchar(300) NOT NULL,
              PRIMARY KEY (`Sequence`)
            ) ENGINE=InnoDB DEFAULT CHARSET=latin1;""",
            """ALTER TABLE `trna` DROP COLUMN `Structure`;""",
            """UPDATE `version` SET `SchemaVersion` = 8;"""
            ],
        step_summary_dict: {
            inaccurate_column: [
                "trna.Product",
                "trna.Sequence",
                "trna.InfernalScore",
                "trna.AminoAcid"
                ],
            lost_column: [
                "trna.Source",
                "trna.Structure",
                "trna.Name",
                "tmrna.Name",
                "tmrna.Length"
                ]
            }
        },

    "downgrade_10_to_9": {
        statements: [
            """UPDATE `gene` SET `Translation` = NULL WHERE LENGTH(`Translation`) > 5000;""",
            """ALTER TABLE `gene` MODIFY COLUMN `Translation` varchar(5000) DEFAULT NULL;""",
            """UPDATE `version` SET `SchemaVersion` = 9;"""
            ],
        step_summary_dict: {
            lost_data: ["gene.Translation"]
            }
        }
    }



"""Functions for running rpsblast and parsing the output."""

from Bio.Blast import NCBIXML

from pdm_utils.functions.fasta import write_fasta
from pdm_utils.functions.subprocess import run_command
from pdm_utils.functions.multiprocess import parallelize
from pdm_utils.classes.error import BlastError

BATCH_SIZE = 250


def get_version(program="rpsblast", long=False):
    """Get the version string from BLAST.

    :param program: globally executable rpsblast(+) to get version for
    :type program: str or pathlib.Path
    :param long: return the long version string
    :type long: bool
    :return: version
    """
    command = f"{program} -version"

    try:
        stderr, stdout = run_command(command)
        if stdout:
            raise BlastError(stdout)
        if long:
            version = stderr.split("\n")[1].split("blast ")[-1]
        else:
            version = stderr.split("\n")[0].split(": ")[-1]
    except FileNotFoundError:
        version = None

    return version


def rpsblast(program, query, db, out, evalue, comp_based_stats=1,
             seg="no", outfmt=5):
    """Run rpsblast(+).

    Query the sequences in `infile` against `database` with results
    written to `outfile`.

    :param program: name or path of the rpsblast(+) program to call
    :type program: str or pathlib.Path
    :param query: path to input file containing query sequences
    :type query: str or pathlib.Path
    :param db: path to database containing NCBI conserved domains
    :type db: str or pathlib.Path
    :param out: path to output file to write results to
    :type out: str or pathlib.Path
    :param evalue: significance threshold to retain Cdd hits
    :type evalue: int or float
    :param comp_based_stats: adjust stats for low-complexity regions
    :type comp_based_stats: int
    :param seg: filter query sequence with SEG
    :type seg: str
    :param outfmt: output file options
    :type outfmt: int or str
    :return: out
    :raise: BlastError
    """
    command = f"{program} -query {query} -db {db} -out {out} -outfmt " \
              f"{outfmt} -evalue {evalue} -seg {seg} -comp_based_stats " \
              f"{comp_based_stats}"

    stdout, stderr = run_command(command)
    if stderr:
        # Problem with filepaths or other parameters
        raise BlastError(stderr)

    return out


def blastall(program, sequences, db, tmp_dir, evalue, cpus=1,
             batch_size=BATCH_SIZE):
    """Query each of the given sequences against the NCBI conserved
    domain database using rpsblast.

    Batches are set up to be reasonably load balancing. The more
    sequences are searched, the better the load balancing works.

    :param program: name or path of the rpsblast(+) program to call
    :type program: str or pathlib.Path
    :param sequences: the (geneid, translation) pairs to search
    :type sequences: list[tuple[str, str]]
    :param db: path to database containing NCBI conserve domains
    :type db: str or pathlib.Path
    :param tmp_dir: directory to use for temporary files
    :type tmp_dir: pathlib.Path
    :param evalue: significance threshold to retain Cdd hits
    :type evalue: int or float
    :param batch_size: maximum number of genes to run per batch
    :type batch_size: int
    :param cpus: number of CPU cores to use
    :type cpus: int
    :return: outfiles
    :rtype: list[str or pathlib.Path]
    """
    # Determine appropriate number of partitions for good load balancing
    num_batches = max((len(sequences) // batch_size, 1))
    while len(sequences) / num_batches > batch_size or num_batches % cpus != 0:
        num_batches += 1

    # Sort the sequences by translation length
    sequences = sorted(sequences, key=lambda x: len(x[-1]))

    jobs = list()
    for batch_i in range(num_batches):
        geneids, translations = list(), list()
        for batch_index in range(len(sequences))[batch_i::num_batches]:
            geneid, translation = sequences[batch_index]
            geneids.append(geneid)
            translations.append(translation)

        infile = tmp_dir.joinpath(f"rpsblast_batch_{batch_i}.fasta")
        outfile = tmp_dir.joinpath(f"rpsblast_batch_{batch_i}.xml")

        write_fasta(geneids, translations, infile)

        jobs.append((program, infile, db, outfile, evalue))

    outfiles = parallelize(jobs, cpus, rpsblast, verbose=True)

    return outfiles


def parse_xml(filepath, sequences, evalue):
    """Return conserved domain hits parsed from an XML results file
    output by rpsblast.

    :param filepath: the rpsblast XML output file
    :type filepath: pathlib.Path
    :param sequences: the geneid, translation pairs queried
    :type sequences: list[tuple[str, str]]
    :param evalue: significance threshold to retain HSPs
    :type evalue: float
    :return: domain_data
    """
    sequence_map = dict()
    for geneid, translation in sequences:
        sequence_map[geneid] = translation

    domain_data = dict()
    with open(filepath, "r") as blast_reader:
        for blast_record in NCBIXML.parse(blast_reader):
            query, domains = sequence_map[blast_record.query], list()
            for alignment in blast_record.alignments:
                domain_id, name, description = None, None, None

                definition = alignment.hit_def.replace("\"", "\'")
                def_list = definition.split(",")
                if len(def_list) == 1:
                    description = def_list[0].strip()
                else:
                    domain_id = def_list[0].strip()
                    name = def_list[1].strip()

                    if len(def_list) > 2:
                        description = ",".join(def_list[2:]).strip()

                for hsp in alignment.hsps:
                    if hsp.expect >= evalue:
                        continue

                    domain = {"HitID": alignment.hit_id,
                              "DomainID": domain_id,
                              "Name": name,
                              "Description": description,
                              "Expect": float(hsp.expect),
                              "QueryStart": int(hsp.query_start),
                              "QueryEnd": int(hsp.query_end)}
                    domains.append(domain)

            domain_data[query] = domains

    return domain_data



"""Pipeline for mapping the differences of PhamIDs between dataabse"""
import argparse
import shutil
import sys
import time
from pathlib import Path

from pdm_utils.functions import basic
from pdm_utils.functions import configfile
from pdm_utils.functions import fileio
from pdm_utils.functions import pipelines_basic

# GLOBAL VARIABLES
# -----------------------------------------------------------------------------
DEFAULT_FOLDER_NAME = f"{time.strftime('%Y%m%d')}_pham_finder"
DEFAULT_FOLDER_PATH = Path.cwd()

PHAM_FINDER_HEADER = ["Reference Pham", "Corresponding Phams"]


def main(unparsed_args_list):
    """Uses parsed args to run the entirety of the pham_finder pipeline.

    :param unparsed_args_list: Input a list of command line args.
    :type unparsed_args_list: list[str]
    """
    args = parse_pham_finder(unparsed_args_list)

    config = configfile.build_complete_config(args.config_file)

    alchemist = pipelines_basic.build_alchemist(None, ask_database=False,
                                                config=config)
    
    values = None
    if args.input:
        values = pipelines_basic.parse_value_input(args.input)

    execute_pham_finder(alchemist, args.folder_path, args.folder_name,
                        args.adatabase, args.bdatabase, values=values,
                        filters=args.filters, groups=args.groups, 
                        sort=args.sort, show_per=args.show_percentages,
                        use_locus=args.use_locus, verbose=args.verbose)


def parse_pham_finder(unparsed_args_list):
    """Parses pham_finder arguments and stores them with an argparse object.

    :param unparsed_args_list: Input a list of command line args.
    :type unparsed_args_list: list[str]
    :returns: ArgParse module parsed args.
    """
    A_DATABASE_HELP = """
        Name of the MySQL database to retrieve reference phams from.
        """
    B_DATABASE_HELP = """
        Name of the MySQL database to find phams for.
        """

    CONFIG_FILE_HELP = """
        Find option that enables use of a config file for sourcing credentials
            Follow selection argument with the path to the config file
            specifying MySQL and NCBI credentials.
        """
    VERBOSE_HELP = """
        Export option that enables progress print statements.
        """
    FOLDER_PATH_HELP = """
        Export option to change the path
        of the directory where the exported files are stored.
            Follow selection argument with the path to the
            desired export directory.
        """
    FOLDER_NAME_HELP = """
        Export option to change the name
        of the directory where the exported files are stored.
            Follow selection argument with the desired name.
        """

    IMPORT_FILE_HELP = """
        Selection input option that imports values from a csv file.
            Follow selection argument with path to the
            csv file containing the names of each genome in the first column.
        """
    SINGLE_PHAMIDS_HELP = """
        Selection input option that imports values from cmd line input.
            Follow selection argument with space separated
            names of genomes in the database.
        """
    WHERE_HELP = """
        Data filtering option that filters data by the inputted expressions.
            Follow selection argument with formatted filter expression:
                {Table}.{Column}={Value}
        """
    GROUP_BY_HELP = """
        Data selection option that groups data by the inputted columns.
            Follow selection argument with formatted column expressions:
                {Table}.{Column}={Value}
        """
    ORDER_BY_HELP = """
        Data selection option that sorts data by the inputted columns.
            Follow selection argument with formatted column expressions:
                {Table}.{Column}={Value}
        """

    SHOW_PERCENTAGES_HELP = """
        Pham finder option that enables the display of the percent of 
        the referenced pham that the found pham covers.
        """
    USE_LOCUS_HELP = """
        Pham finder option that converts between phams using LocusTags.
        """

    parser = argparse.ArgumentParser()
    parser.add_argument("adatabase", type=str, 
                        help=A_DATABASE_HELP)
    parser.add_argument("bdatabase", type=str, 
                        help=B_DATABASE_HELP)

    parser.add_argument("-c", "--config_file", 
                        type=pipelines_basic.convert_file_path,
                        help=CONFIG_FILE_HELP)
    parser.add_argument("-m", "--folder_name", type=str,
                        help=FOLDER_NAME_HELP)
    parser.add_argument("-o", "--folder_path", 
                        type=pipelines_basic.convert_dir_path,
                        help=FOLDER_PATH_HELP)
    parser.add_argument("-v", "--verbose", action="store_true",
                        help=VERBOSE_HELP)

    parser.add_argument("-if", "--import_file", dest="input",
                        type=pipelines_basic.convert_file_path,
                        help=IMPORT_FILE_HELP)
    parser.add_argument("-in", "--import_names", dest="input", nargs="*",
                        help=SINGLE_PHAMIDS_HELP)
    parser.add_argument("-f", "--where", nargs="?", dest="filters",
                        help=WHERE_HELP)
    parser.add_argument("-g", "--group_by", nargs="*", dest="groups",
                        help=GROUP_BY_HELP)
    parser.add_argument("-s", "--order_by", nargs="*", dest="sort",
                        help=ORDER_BY_HELP)

    parser.add_argument("-per", "--show_percentages", action="store_true", 
                        help=SHOW_PERCENTAGES_HELP)
    parser.add_argument("-ul", "--use_locus", action="store_true",
                        help=USE_LOCUS_HELP)

    parser.set_defaults(folder_name=DEFAULT_FOLDER_NAME,
                        folder_path=DEFAULT_FOLDER_PATH,
                        config_file=None, verbose=False, input=[],
                        filters="", groups=[], sort=[],
                        show_percentages=False)

    parsed_args = parser.parse_args(unparsed_args_list[2:])
    return parsed_args


# TODO Owen Needs unittests
def execute_pham_finder(alchemist, folder_path, folder_name,
                        adatabase, bdatabase, values=None,
                        filters="", groups=[], sort=[],
                        show_per=False, use_locus=False, verbose=False):
    """Executes the entirety of the file export pipeline.

    :param alchemist: A connected and fully build AlchemyHandler object.
    :type alchemist: AlchemyHandler
    :param folder_path: Path to a valid dir for new dir creation.
    :type folder_path: Path
    :param folder_name: A name for the export folder.
    :type folder_name: str
    :param adatabase: Name of reference database to source phams from.
    :type adatabase: str
    :param bdatabase: Name of database to find corresponding phams for.
    :type bdatabase: str
    :param values: List of values to filter database results:
    :type values: list[str]
    :param verbose: A boolean value to toggle progress print statements.
    :type verbose: bool
    :param table: MySQL table name.
    :type table: str
    :param filters: A list of lists with filter values, grouped by ORs.
    :type filters: str
    :param groups: A list of supported MySQL column names to group by.
    :type groups: A list of supported MySQL column names to group by.
    :type groups: list[str]
    :param sort: A list of supported MySQL column names to sort by.
    :type sort: list[str]
    :param show_per: Enables display gene coverage of the corresponding phams.
    :type show_per: bool
    :param use_locus: Toggles conversion between phams using LocusTag instead
    :type use_locus: bool
    """
    if not (adatabase in alchemist.databases and
            bdatabase in alchemist.databases):
        print("User credentials does not have access to both "
              f"databases {adatabase} and {bdatabase}.\n"
              "Please check your database access and try again.")
        sys.exit(1)

    alchemist.database = adatabase
    alchemist.connect()
    a_filter = pipelines_basic.build_filter(alchemist, "gene.PhamID", filters,
                                            values=values, verbose=verbose)

    alchemist.database = bdatabase
    alchemist.connect()
    if use_locus:
        b_filter = pipelines_basic.build_filter(alchemist, "gene.LocusTag", "")
    else:
        b_filter = pipelines_basic.build_filter(alchemist, "gene", "")

    if sort:
        try:
            a_filter.sort(sort)
        except:
            print("Please check your syntax for sorting columns:\n"
                  f"{', '.join(sort)}")
            sys.exit(1)

    if verbose:
        print("Creating pham_finder folder...")
    export_path = folder_path.joinpath(folder_name)
    export_path = basic.make_new_dir(folder_path, export_path, attempt=50)

    conditionals_map = pipelines_basic.build_groups_map(a_filter, export_path,
                                                        groups=groups,
                                                        verbose=verbose)

    if verbose:
        print("Prepared query and path structure, beginning export...")

    values = a_filter.values
    for mapped_path in conditionals_map.keys():
        a_filter.reset()
        a_filter.values = values

        conditionals = conditionals_map[mapped_path]
        a_filter.values = a_filter.build_values(where=conditionals)

        if a_filter.hits() == 0:
            print("No database entries received from gene.PhamID "
                  f"for '{mapped_path}'.")
            shutil.rmtree(mapped_path)
            continue

        if sort:
            a_filter.sort(sort)

        mapped_phams = find_phams(a_filter, b_filter, show_per=show_per)
        if not mapped_phams:
            print("Phams are consistent between the two databases "
                  f"for '{mapped_path}'.")
            shutil.rmtree(mapped_path)
            continue

        out_data_dicts = []
        for ref_pham, corr_phams in mapped_phams.items():
            data_dict = {}
            data_dict[PHAM_FINDER_HEADER[0]] = ref_pham
            data_dict[PHAM_FINDER_HEADER[1]] = corr_phams
            out_data_dicts.append(data_dict)

        file_path = mapped_path.joinpath("PhamMap.csv")
        fileio.export_data_dict(out_data_dicts, file_path, PHAM_FINDER_HEADER,
                                include_headers=True)


# TODO Owen Needs unittests
def find_phams(a_filter, b_filter, show_per=False, use_locus=False):
    """Find phams helper function that finds phams via GeneID intermediates.

    :param a_filter: Fully built Filter connected to the reference database.
    :type a_filter: Filter
    :param b_filter: Fully build Filter connected to a database.
    :type b_filter: Filter
    :param show_per: Enables display gene coverage of the corresponding phams.
    :type show_per: bool
    :returns: Returns a dictionary mapping original phams to corresponding phams
    :rtype: dict{int:str}
    """
    if use_locus:
        genes = a_filter.retrieve("gene.GeneID")
    else:
        genes = a_filter.retrieve("gene.GeneID")

    mapped_phams = {}
    for pham, data_dict in genes.items():
        if use_locus:
            values = data_dict["LocusTag"]
            values.remove(None)
            b_filter.values = values
        else:
            b_filter.values = data_dict["GeneID"]        

        if show_per:
            pham_groups = b_filter.group("gene.PhamID")
            phams_list = list(pham_groups.keys())
            total_genes = 0
            
            for grouped_genes in pham_groups.values():
                total_genes += len(grouped_genes)
        else:
            phams_list = b_filter.transpose("gene.PhamID")

        if len(phams_list) == 1:
            if phams_list[0] == pham:
                print(f"{phams_list[0]} == {pham}")
                continue
            else:
                corr_phams = str(phams_list[0])
        elif len(phams_list) == 0:
            corr_phams = "None"
        else:
            if show_per:
                for i in range(len(phams_list)):
                    join_pham = phams_list[i]
                    percent = (len(pham_groups[join_pham])\
                                  /total_genes) * 100
                    percent = round(percent, 1)
                    phams_list[i] = "".join([str(join_pham), 
                                        "(", str(percent), "%)"])
            corr_phams = ";".join([str(join_pham) for join_pham in phams_list])

        mapped_phams[pham] = corr_phams
        
    return mapped_phams


if __name__ == "__main__":
    main(sys.argv)



"""Functions to interact with MySQL."""

import getpass
import subprocess
import sys
import re

from Bio.Alphabet import IUPAC
from Bio.Seq import Seq
import sqlalchemy

from pdm_utils.classes import cds, trna, tmrna
from pdm_utils.classes import genome
from pdm_utils.classes import genomepair
from pdm_utils.constants import constants
from pdm_utils.functions import basic
from pdm_utils.functions import mysqldb_basic

def parse_phage_table_data(data_dict, trans_table=11, gnm_type=""):
    """Parse a MySQL database dictionary to create a Genome object.

    :param data_dict:
        Dictionary of data retrieved from the phage table.
    :type data_dict: dict
    :param trans_table:
        The translation table that can be used to translate CDS features.
    :type trans_table: int
    :param gnm_type: Identifier for the type of genome.
    :type gnm_type: str
    :returns: A pdm_utils genome object.
    :rtype: genome
    """

    gnm = genome.Genome()
    try:
        gnm.id = data_dict["PhageID"]
    except:
        pass

    try:
        gnm.accession = data_dict["Accession"]
    except:
        pass

    try:
        gnm.name = data_dict["Name"]
    except:
        pass

    try:
        gnm.host_genus = data_dict["HostGenus"]
    except:
        pass

    try:
        # Sequence data is stored as MEDIUMBLOB, so decode to string.
        gnm.set_sequence(data_dict["Sequence"].decode("utf-8"))
    except:
        pass

    try:
        gnm.length = int(data_dict["Length"])
    except:
        pass

    try:
        # DateLastModified gets returned as a datetime.datetime object.
        # TODO some phages have no date, so it will be returned NULL.
        gnm.date = data_dict["DateLastModified"]
    except:
        pass

    try:
        gnm.description = data_dict["Notes"].decode("utf-8")
    except:
        pass

    try:
        gnm.gc = float(data_dict["GC"])
    except:
        pass

    try:
        # Singletons are stored in the MySQL database as NULL, which gets
        # returned as None.
        gnm.set_cluster(data_dict["Cluster"])
    except:
        pass


    # Non-subclustered phages are stored in the MySQL database as NULL, which gets
    # returned as None.
    try:
        gnm.set_subcluster(data_dict["Subcluster"])
    except:
        pass

    try:
        gnm.annotation_status = data_dict["Status"]
    except:
        pass

    try:
        gnm.set_retrieve_record(data_dict["RetrieveRecord"])
    except:
        pass

    try:
        gnm.set_annotation_author(data_dict["AnnotationAuthor"])
    except:
        pass

    gnm.translation_table = trans_table
    gnm.type = gnm_type
    return gnm


def parse_gene_table_data(data_dict, trans_table=11):
    """Parse a MySQL database dictionary to create a Cds object.

    :param data_dict:
        Dictionary of data retrieved from the gene table.
    :type data_dict: dict
    :param trans_table:
        The translation table that can be used to translate CDS features.
    :type trans_table: int
    :returns: A pdm_utils Cds object.
    :rtype: Cds
    """

    ftr = cds.Cds()
    ftr.type = "CDS"
    try:
        ftr.id = data_dict["GeneID"]
    except:
        pass

    try:
        ftr.genome_id = data_dict["PhageID"]
    except:
        pass

    try:
        ftr.start = int(data_dict["Start"])
    except:
        pass

    try:
        ftr.stop = int(data_dict["Stop"])
    except:
        pass

    try:
        ftr.parts = int(data_dict["Parts"])
    except:
        pass

    ftr.coordinate_format = "0_half_open"

    try:
        ftr.length = int(data_dict["Length"])
    except:
        pass

    try:
        ftr.name = data_dict["Name"]
    except:
        pass

    try:
        ftr.set_translation(data_dict["Translation"].decode("utf-8"))
    except:
        pass

    try:
        ftr.orientation = data_dict["Orientation"]
    except:
        pass

    try:
        ftr.description = data_dict["Notes"].decode("utf-8")
    except:
        pass

    try:
        ftr.set_locus_tag(data_dict["LocusTag"])
    except:
        pass

    try:
        ftr.pham_id = int(data_dict["PhamID"])
    except:
        pass

    try:
        ftr.domain_status = int(data_dict["DomainStatus"])
    except:
        pass

    try:
        ftr.translation_table = trans_table
    except:
        pass

    return ftr


# TODO Christian review
def parse_trna_table_data(data_dict):
    """Parse a MySQL database dictionary to create a Trna object.

    :param data_dict:
        Dictionary of data retrieved from the gene table.
    :type data_dict: dict
    :returns: A pdm_utils Trna object.
    :rtype: Trna
    """

    ftr = trna.Trna()
    ftr.type = "tRNA"

    try:
        ftr.id = data_dict["GeneID"]
    except:
        pass

    try:
        ftr.genome_id = data_dict["PhageID"]
    except:
        pass

    try:
        ftr.start = int(data_dict["Start"])
    except:
        pass

    try:
        ftr.stop = int(data_dict["Stop"])
    except:
        pass

    try:
        ftr.length = int(data_dict["Length"])
    except:
        pass

    try:
        ftr.name = data_dict["Name"]
    except:
        pass

    try:
        ftr.orientation = data_dict["Orientation"]
    except:
        pass

    try:
        ftr.set_locus_tag(data_dict["LocusTag"])
    except:
        pass

    try:
        ftr.note = data_dict["Note"].decode("utf-8")
    except:
        pass

    try:
        ftr.amino_acid = data_dict["AminoAcid"]
    except:
        pass

    try:
        ftr.anticodon = data_dict["Anticodon"]
    except:
        pass

    try:
        ftr.structure = data_dict["Structure"].decode("utf-8")
    except:
        pass

    try:
        ftr.use = data_dict["Source"]
    except:
        pass

    try:
        ftr.parts = 1
    except:
        pass

    ftr.coordinate_format = "0_half_open"

    return ftr

# TODO Christian review
def parse_tmrna_table_data(data_dict):
    """Parse a MySQL database dictionary to create a Tmrna object.

    :param data_dict:
        Dictionary of data retrieved from the gene table.
    :type data_dict: dict
    :returns: A pdm_utils Tmrna object.
    :rtype: Tmrna
    """

    ftr = tmrna.Tmrna()
    ftr.type = "tmRNA"

    try:
        ftr.id = data_dict["GeneID"]
    except:
        pass

    try:
        ftr.genome_id = data_dict["PhageID"]
    except:
        pass

    try:
        ftr.start = int(data_dict["Start"])
    except:
        pass

    try:
        ftr.stop = int(data_dict["Stop"])
    except:
        pass

    try:
        ftr.length = int(data_dict["Length"])
    except:
        pass

    try:
        ftr.name = data_dict["Name"]
    except:
        pass

    try:
        ftr.orientation = data_dict["Orientation"]
    except:
        pass

    try:
        ftr.set_locus_tag(data_dict["LocusTag"])
    except:
        pass

    try:
        ftr.note = data_dict["Note"].decode("utf-8")
    except:
        pass

    try:
        ftr.peptide_tag = data_dict["PeptideTag"]
    except:
        pass

    try:
        ftr.parts = 1
    except:
        pass

    ftr.coordinate_format = "0_half_open"

    return ftr


def parse_feature_data(engine, ftr_type, column=None, phage_id_list=None,
                       query=None):
    """Returns Cds objects containing data parsed from a
    MySQL database.

    :param engine:
        This parameter is passed directly to the 'retrieve_data' function.
    :type engine: Engine
    :param query:
        This parameter is passed directly to the 'retrieve_data' function.
    :type query: str
    :param ftr_type:
        Indicates the type of features retrieved.
    :type ftr_type: str
    :param column:
        This parameter is passed directly to the 'retrieve_data' function.
    :type column: str
    :param phage_id_list:
        This parameter is passed directly to the 'retrieve_data' function.
    :type phage_id_list: list
    :returns: A list of pdm_utils Cds objects.
    :rtype: list
    """
    result_list = mysqldb_basic.retrieve_data(engine, column=column,
                                              query=query,
                                              id_list=phage_id_list)
    ftrs = []
    for data_dict in result_list:
        if ftr_type == "cds":
            ftr = parse_gene_table_data(data_dict)
        elif ftr_type == "trna":
            ftr = parse_trna_table_data(data_dict)
        elif ftr_type == "tmrna":
            ftr = parse_tmrna_table_data(data_dict)
        else:
            # If the ftr_type is invalid, just take the data dictionary.
            # Alternatively it could raise an error.
            ftr = data_dict
        ftrs.append(ftr)
    return ftrs


def parse_genome_data(engine, phage_id_list=None, phage_query=None,
                      gene_query=None, trna_query=None, tmrna_query=None,
                      gnm_type=""):
    """Returns a list of Genome objects containing data parsed from a MySQL
    database.

    :param engine:
        This parameter is passed directly to the 'retrieve_data' function.
    :type engine: Engine
    :param phage_query:
        This parameter is passed directly to the 'retrieve_data' function
        to retrieve data from the phage table.
    :type phage_query: str
    :param gene_query:
        This parameter is passed directly to the 'parse_feature_data' function
        to retrieve data from the gene table.
        If not None, pdm_utils Cds objects for all of the phage's
        CDS features in the gene table will be constructed
        and added to the Genome object.
    :type gene_query: str
    :param trna_query:
        This parameter is passed directly to the 'parse_feature_data' function
        to retrieve data from the trna table.
        If not None, pdm_utils Trna objects for all of the phage's
        tRNA features in the trna table will be constructed
        and added to the Genome object.
    :type trna_query: str
    :param tmrna_query:
        This parameter is passed directly to the 'parse_feature_data' function
        to retrieve data from the tmrna table.
        If not None, pdm_utils Tmrna objects for all of the phage's
        tmRNA features in the tmrna table will be constructed
        and added to the Genome object.
    :type tmrna_query: str
    :param phage_id_list:
        This parameter is passed directly to the 'retrieve_data' function.
        If there is at at least one valid PhageID, a pdm_utils genome
        object will be constructed only for that phage. If None, or an
        empty list,  genome objects for all phages in the
        database will be constructed.
    :type phage_id_list: list
    :param gnm_type: Identifier for the type of genome.
    :type gnm_type: str
    :returns: A list of pdm_utils Genome objects.
    :rtype: list
    """
    COLUMN = "PhageID"
    genome_list = []
    result_list1 = mysqldb_basic.retrieve_data(engine, column=COLUMN,
                                               id_list=phage_id_list,
                                               query=phage_query)
    for data_dict in result_list1:
        gnm = parse_phage_table_data(data_dict, gnm_type=gnm_type)

        if gene_query is not None:
            cds_list = parse_feature_data(engine, "cds", column=COLUMN,
                                          phage_id_list=[gnm.id],
                                          query=gene_query)

            for x in range(len(cds_list)):
                cds_list[x].genome_length = gnm.length
            gnm.cds_features = cds_list

        if trna_query is not None:
            trna_list = parse_feature_data(engine, "trna", column=COLUMN,
                                          phage_id_list=[gnm.id],
                                          query=trna_query)

            for x in range(len(trna_list)):
                trna_list[x].genome_length = gnm.length
            gnm.trna_features = trna_list

        if tmrna_query is not None:
            tmrna_list = parse_feature_data(engine, "tmrna", column=COLUMN,
                                          phage_id_list=[gnm.id],
                                          query=tmrna_query)

            for x in range(len(tmrna_list)):
                tmrna_list[x].genome_length = gnm.length
            gnm.tmrna_features = tmrna_list

        genome_list.append(gnm)
    return genome_list


def create_seq_set(engine):
    """Create set of genome sequences currently in a MySQL database.

    :param engine: SQLAlchemy Engine object able to connect to a MySQL database.
    :type engine: Engine
    :returns: A set of unique values from phage.Sequence.
    :rtype: set
    """
    query = "SELECT Sequence FROM phage"

    # Returns a list of items, where each item is a tuple of
    # SQL data for each row in the table.
    result_list = engine.execute(query).fetchall()

    # Convert to a set of sequences.
    # Sequence data is stored as MEDIUMBLOB, so data is returned as bytes
    # "b'AATT", "b'TTCC", etc.
    result_set = set()
    for tup in result_list:
        gnm_seq = tup[0].decode("utf-8")
        gnm_seq = Seq(gnm_seq, IUPAC.ambiguous_dna).upper()
        result_set.add(gnm_seq)
    return result_set


def create_update(table, field2, value2, field1, value1):
    """Create MySQL UPDATE statement.

    "'UPDATE <table> SET <field2> = '<value2' WHERE <field1> = '<data1>'."

    When the new value to be added is 'singleton' (e.g. for Cluster
    fields), or an empty value (e.g. None, "none", etc.),
    the new value is set to NULL.

    :param table: The database table to insert information.
    :type table: str
    :param field1: The column upon which the statement is conditioned.
    :type field1: str
    :param value1:
        The value of 'field1' upon which the statement is conditioned.
    :type value1: str
    :param field2: The column that will be updated.
    :type field2: str
    :param value2:
        The value that will be inserted into 'field2'.
    :type value2: str
    :returns: A MySQL UPDATE statement.
    :rtype: set
    """
    check_set = constants.EMPTY_SET | {"Singleton"}
    part1 = f"UPDATE {table} SET {field2} = "
    part3 = f" WHERE {field1} = '{value1}';"
    part2 = mysqldb_basic.convert_for_sql(value2, check_set=check_set, single=True)
    statement = part1 + part2 + part3
    return statement


def create_delete(table, field, data):
    """Create MySQL DELETE statement.

    "'DELETE FROM <table> WHERE <field> = '<data>'."

    :param table: The database table to insert information.
    :type table: str
    :param field: The column upon which the statement is conditioned.
    :type field: str
    :param data: The value of 'field' upon which the statement is conditioned.
    :type data: str
    :returns: A MySQL DELETE statement.
    :rtype: str
    """
    statement = f"DELETE FROM {table} WHERE {field} = '{data}';"
    return statement


def create_gene_table_insert(cds_ftr):
    """Create a MySQL gene table INSERT statement.

    :param cds_ftr: A pdm_utils Cds object.
    :type cds_ftr: Cds
    :returns:
        A MySQL statement to INSERT a new row in the 'gene' table
        with data for several fields.
    :rtype: str
    """
    locus_tag = mysqldb_basic.convert_for_sql(cds_ftr.locus_tag,
                                              check_set={""}, single=False)

    # cds_ftr.translation is a BioPython Seq object.
    # It is coerced to string by default.
    # Statement should be triple quoted.
    # Descriptions may contain a "'", so at least description attribute needs
    # to be encapsulated with double quotes.
    # locus_tag can be NULL or can be a string, so it should not be
    # encapsulated with '"'.
    statement = ("""INSERT INTO gene """
                 """(GeneID, PhageID, Start, Stop, Length, Name, """
                 """Translation, Orientation, Notes, LocusTag, Parts) """
                 """VALUES """
                 """("{}", "{}", {}, {}, {}, "{}", "{}", "{}", "{}", {}, {});""")
    statement = statement.format(cds_ftr.id, cds_ftr.genome_id,
                                 cds_ftr.start, cds_ftr.stop,
                                 cds_ftr.length,
                                 cds_ftr.name, cds_ftr.translation,
                                 cds_ftr.orientation, cds_ftr.description,
                                 locus_tag, cds_ftr.parts)
    return statement


def create_trna_table_insert(trna_ftr):
    """
    Create a MySQL trna table INSERT statement.
    :param trna_ftr: a pdm_utils Trna object
    :type trna_ftr: Trna
    :returns: a MySQL statement to INSERT a new row in the 'trna' table
    with all of trna_ftr's relevant data
    :rtype: str
    """
    # Any values that could be none should be run through mysqldb_basic's
    # convert_for_sql method so that the statement is does not cause errors
    # when evaluated
    geneid, phageid = trna_ftr.id, trna_ftr.genome_id
    name = trna_ftr.name
    locus_tag = mysqldb_basic.convert_for_sql(
        trna_ftr.locus_tag, check_set={""}, single=False)
    start, stop, length = trna_ftr.start, trna_ftr.stop, trna_ftr.length
    orientation = trna_ftr.orientation
    note = mysqldb_basic.convert_for_sql(
        trna_ftr.note, check_set={""}, single=False)
    amino_acid, anticodon = trna_ftr.amino_acid, trna_ftr.anticodon
    structure = mysqldb_basic.convert_for_sql(
        trna_ftr.structure, check_set={""}, single=False)
    source = mysqldb_basic.convert_for_sql(
        trna_ftr.use, check_set={None}, single=False)

    statement = ("""INSERT INTO trna """
                 """(GeneID, PhageID, Start, Stop, Length, """
                 """Name, Orientation, Note, LocusTag, AminoAcid, Anticodon, """
                 """Structure, Source) VALUES ("{}", "{}", {}, {}, {}, """
                 """"{}", "{}", {}, {}, "{}", "{}", {}, {});""")
    statement = statement.format(geneid, phageid, start, stop, length,
                                 name, orientation, note, locus_tag,
                                 amino_acid, anticodon, structure, source)
    return statement


def create_tmrna_table_insert(tmrna_ftr):
    """

    :param tmrna_ftr:
    :return:
    """
    geneid, phageid = tmrna_ftr.id, tmrna_ftr.genome_id
    name = tmrna_ftr.name
    locus_tag = mysqldb_basic.convert_for_sql(
        tmrna_ftr.locus_tag, check_set={""}, single=False)
    start, stop, length = tmrna_ftr.start, tmrna_ftr.stop, tmrna_ftr.length
    orientation = tmrna_ftr.orientation
    note = mysqldb_basic.convert_for_sql(
        tmrna_ftr.note, check_set={""}, single=False)
    peptide_tag = mysqldb_basic.convert_for_sql(
        tmrna_ftr.peptide_tag, check_set={""}, single=False)

    statement = ("""INSERT INTO tmrna """
                 """(GeneID, PhageID, Start, Stop, Length, """
                 """Name, Orientation, Note, LocusTag, PeptideTag) VALUES """
                 """("{}", "{}", {}, {}, {}, "{}", "{}", {}, {}, {});""")
    statement = statement.format(geneid, phageid, start, stop, length,
                                 name, orientation, note, locus_tag,
                                 peptide_tag)
    return statement


def create_phage_table_insert(gnm):
    """Create a MySQL phage table INSERT statement.

    :param gnm: A pdm_utils Genome object.
    :type gnm: Genome
    :returns:
        A MySQL statement to INSERT a new row in the 'phage' table
        with data for several fields.
    :rtype: str
    """
    cluster = mysqldb_basic.convert_for_sql(gnm.cluster,
                                            check_set={"Singleton"},
                                            single=True)
    subcluster = mysqldb_basic.convert_for_sql(gnm.subcluster,
                                               check_set={"none"},
                                               single=True)

    # gnm.seq is a BioPython Seq object.
    # It is coerced to string by default.
    statement = ("""INSERT INTO phage """
                 """(PhageID, Accession, Name, HostGenus, Sequence, """
                 """Length, GC, Status, DateLastModified, RetrieveRecord, """
                 """AnnotationAuthor, Cluster, Subcluster) """
                 """VALUES """
                 """('{}', '{}', '{}', '{}', '{}', {}, {}, '{}', """
                 """'{}', {}, {}, {}, {});""")
    statement = statement.format(
                    gnm.id, gnm.accession, gnm.name, gnm.host_genus, gnm.seq,
                    gnm.length, gnm.gc, gnm.annotation_status, gnm.date,
                    gnm.retrieve_record, gnm.annotation_author, cluster,
                    subcluster)

    return statement


def create_genome_statements(gnm, tkt_type=""):
    """Create list of MySQL statements based on the ticket type.

    :param gnm: A pdm_utils Genome object.
    :type gnm: Genome
    :param tkt_type: 'add' or 'replace'.
    :type tkt_type: str
    :returns:
        List of MySQL statements to INSERT all data
        from a genome into the database
        (DELETE FROM genome, INSERT INTO phage, INSERT INTO gene, ...).
    :rtype: list
    """

    stmts = []
    if tkt_type == "replace":
        stmt1 = create_delete("phage", "PhageID", gnm.id)
        stmts.append(stmt1)
    stmt2 = create_phage_table_insert(gnm)
    stmts.append(stmt2)
    for cds_ftr in gnm.cds_features:
        stmt3 = create_gene_table_insert(cds_ftr)
        stmts.append(stmt3)
    for trna_ftr in gnm.trna_features:
        stmt4 = create_trna_table_insert(trna_ftr)
        stmts.append(stmt4)
    for tmrna_ftr in gnm.tmrna_features:
        stmt5 = create_tmrna_table_insert(tmrna_ftr)
        stmts.append(stmt5)

    return stmts



def change_version(engine, amount=1):
    """Change the database version number.

    :param engine: SQLAlchemy Engine object able to connect to a MySQL database.
    :type engine: Engine
    :param amount: Amount to increment/decrement version number.
    :type amount: int
    """
    result = mysqldb_basic.get_first_row_data(engine, "version")
    current = result["Version"]
    new = current + amount
    print(f"Updating version from {current} to {new}.")
    statement = (f"UPDATE version SET Version = {new}")
    engine.execute(statement)



# TODO unittest.
def get_schema_version(engine):
    """Identify the schema version of the database_versions_list.

    Schema version data has not been persisted in every schema version,
    so if schema version data is not found, it is deduced from other
    parts of the schema.

    :param engine: SQLAlchemy Engine object able to connect to a MySQL database.
    :type engine: Engine
    :returns: The version of the pdm_utils database schema.
    :rtype: int
    """
    # 1. If the version table does not exist, schema_version = 0.
    # 2. If there is no schema_version or SchemaVersion field,
    #    it is either schema_version = 1 or 2.
    # 3. If AnnotationAuthor, Program, AnnotationQC, and RetrieveRecord
    #    columns are in phage table, schema_version = 2.


    db_tables = mysqldb_basic.get_tables(engine, engine.url.database)
    if "version" in db_tables:
        version_table = True
    else:
        version_table = False

    if version_table == True:
        version_columns = mysqldb_basic.get_first_row_data(engine, "version")
        if "schema_version" in version_columns.keys():
            schema_version = version_columns["schema_version"]
        elif "SchemaVersion" in version_columns.keys():
            schema_version = version_columns["SchemaVersion"]
        else:
            phage_columns = mysqldb_basic.get_columns(
                                engine, engine.url.database, "phage")
            expected = {"AnnotationAuthor", "Program",
                        "AnnotationQC", "RetrieveRecord"}
            diff = expected - phage_columns
            if len(diff) == 0:
                schema_version = 2
            else:
                schema_version = 1
    else:
        schema_version = 0
    return schema_version

# TODO unittest.
def check_schema_compatibility(engine, pipeline, code_version=None):
    """Confirm database schema is compatible with code.

    If schema version is not compatible, sys.exit is called.

    :param engine: SQLAlchemy Engine object able to connect to a MySQL database.
    :type engine: Engine
    :param pipeline: Description of the pipeline checking compatibility.
    :type pipeline: str
    :param code_version:
        Schema version on which the pipeline operates. If no schema version is
        provided, the package-wide schema version value is used.
    :type code_version: int
    """
    schema_version = get_schema_version(engine)
    if code_version is None:
        code_version = constants.CODE_SCHEMA_VERSION
    if code_version != schema_version:
        print(f"The database schema version is {schema_version}, but "
              f"{pipeline} is compatible with "
              f"schema version {code_version}.")
        sys.exit(1)


# Handles exceptions similar to find_domains pipeline.
# Copied and modeled from MySQLConnectionHandler.execute_transaction,
# but simplified since Engine does the work of connecting to the database.
def execute_transaction(engine, statement_list=[]):
    """Execute list of MySQL statements within a single defined transaction.

    :param engine:
        SQLAlchemy Engine object able to connect to a MySQL databas.
    :type engine: Engine
    :param statement_list:
        a list of any number of MySQL statements with
        no expectation that anything will return
    :returns:
        tuple (result, message)
        WHERE
        result (int) is 0 or 1 status code. 0 means no problems, 1 means problems
        message(str) is a description of the result.
    :rtype: tuple
    """
    msg = "Unable to execute MySQL statements. "
    connection = engine.connect()
    trans = connection.begin()
    try:
        for statement in statement_list:
            connection.execute(statement)
        trans.commit()

    except sqlalchemy.exc.DBAPIError as err:
        err_stmt = err.statement
        sqla_err_type = str(type(err))
        pymysql_err_type = str(type(err.orig))
        pymysql_err_code = err.orig.args[0]
        pymysql_err_msg = err.orig.args[1]
        msg = msg + ("A MySQL error was encountered. "
                     f"SQLAlchemy Error type: {sqla_err_type}. "
                     f"PyMySQL Error type: {pymysql_err_type}. "
                     f"PyMYSQL Error code: {pymysql_err_code}. "
                     f"PyMySQL Error message: {pymysql_err_msg}. "
                     f"Statement: {err_stmt}")
        result = 1
    except TypeError as err:
        msg = msg + "A Python TypeError was encountered. "
        result = 1
    except:
        # TODO not sure how to test this block. Would need to construct a
        # statement that causes a Python built-in exception other than TypeError.
        msg = msg + "An unidentified error was encountered. "
        result = 1
    else:
        msg = "MySQL statements were successfully executed. "
        result = 0

    if result == 1:
        print(msg)
        print("Rolling back transaction...")
        trans.rollback()

    return result, msg

    # Code block below does same thing, but doesn't return a value
    # based on if there is an error.
    # with engine.begin() as connection:
    #     for statement in statement_list:
    #         r1 = connection.execute(statement)




"""Represents a collection of data about a CDS features that are commonly used
to maintain and update SEA-PHAGES phage genomics data.
"""

from collections import OrderedDict
import re

from Bio.Seq import Seq
from Bio.Alphabet import IUPAC
from Bio.SeqFeature import SeqFeature, FeatureLocation, CompoundLocation

from pdm_utils.constants import constants
from pdm_utils.classes import evaluation
from pdm_utils.functions import basic




class Cds:
    """Class to hold data about a CDS feature."""

    def __init__(self):

        # The following attributes are common to any CDS.
        self.id = "" # Gene ID
        self.name = "" # Tends to be an integer for SEA-PHAGES.
        self.seqfeature = None # Biopython SeqFeature object.
        self.start = -1 # Genomic position
        self.stop = -1 # Genomic position
        self.coordinate_format = "" # Indexing format used for coordinates.
        self.orientation = "" #'forward', 'reverse', 'top', 'bottom', etc.
        self.parts = 0 # Number of regions that define the feature
        self.translation_table = 0
        self.translation = Seq("", IUPAC.protein) # Biopython amino acid Seq object.
        self.translation_length = 0
        self.seq = Seq("", IUPAC.ambiguous_dna) # Biopython nucleotide Seq object.
        self.length = 0

        # Information about the genome from which the feature is derived.
        self.genome_id = ""
        self.genome_length = -1


        # The following attributes are common to MySQL database.
        self.pham_id = 0
        self.domain_status = -1
        self.raw_description = "" # Raw gene description
        self.description = "" # Non-generic gene descriptions


        # The following attributes are common to
        # GenBank-formatted flat file records.
        self.locus_tag = "" # Gene ID comprised of PhageID and Gene name
        self._locus_tag_num = "" # Should be digit, but keep as string.
        self.gene = "" # Tends to be an integer, but not guaranteed.
        self.raw_product = ""
        self.raw_function = ""
        self.raw_note = ""
        self.product = ""
        self.function = ""
        self.note = ""
        self._product_num = ""
        self._function_num = ""
        self._note_num = ""
        self._gene_num = ""

        # The following attributes are usefule for processing data
        # from various data sources.
        self.evaluations = []
        self.type = ""
        self._start_stop_orient_id = ()
        self._end_orient_id = ()
        self._start_end_id = ()




    def set_locus_tag(self, tag="", delimiter="_", check_value=None):
        """Set locus tag and parse the locus_tag feature number.

        :param tag:
            Input locus_tag data.
        :type tag: str
        :param delimiter:
            Value used to split locus_tag data.
        :type delimiter: str
        :param check_value:
            Indicates genome name or other value that will be used to parse
            the locus_tag to identify the feature number. If no check_value
            is provided, the genome_id attribute is used.
        :type check_value: str
        """
        # List of delimiters found in locus tags
        # "_" is the most common. The others are rare.
        delimiters = ["_", "-", "."]

        # MySQL database-output format
        if tag is None:
            tag = ""
        self.locus_tag = tag
        if delimiter is None:
            delimiter = basic.choose_most_common(tag, delimiters)
        if check_value is None:
            check_value = self.genome_id
        pattern = re.compile(check_value.lower())
        parts = tag.split(delimiter)

        index = 0
        found = False
        while (index < len(parts) and not found):
            search_result = pattern.search(parts[index].lower())
            if search_result != None:
                found = True
            else:
                index += 1
        if found:
            if index == len(parts) - 1:
                value = parts[index][search_result.end():]
            else:
                value = parts[-1]
        else:
            value = parts[-1]

        # Remove generic 'gp' prefix if present (e.g. TRIXIE_gp10)
        # Sometimes locus tags contain number with character (e.g. TRIXIE_10A)
        if value.lower().startswith("gp"):
            value = value[2:]
        elif value.lower().startswith("orf"):
            value = value[3:]
        else:
            pass

        # If a numeric value can be identified, take it, other wise
        # take the entire value
        left, right = basic.split_string(value)
        if right != "":
            self._locus_tag_num = right
        else:
            self._locus_tag_num = value


    def set_name(self, value=None):
        """Set the feature name.

        Ideally, the name of the CDS will be an integer. This information
        can be stored in multiple fields in the GenBank-formatted flat file.
        The name is derived from one of several qualifiers.

        :param value:
            Indicates a value that should be used to directly set
            the name regardless of the 'gene' and '_locus_tag_num'
            attributes.
        :type value: str
        """
        # The CDS feature number may be stored in several places in the
        # gene, product, note, function, and locus_tag qualifiers of a
        # flat file, but it is unpredictable. Not all of these qualifiers
        # are always present.
        # 1. PECAAN Draft and new SEA-PHAGES Final:
        #    The 'gene' qualifier should be present and contain an integer.
        # 2. SEA-PHAGES Final in GenBank:
        #    The 'gene' qualifier may or may not be present, and
        #    it may or may not have an integer.
        #    The 'locus_tag' qualifier may or may not be present,
        #    and may or may not have an integer.
        # 3. Non-SEA-PHAGES in GenBank:
        #    The 'gene' qualifier may or may not be present, and
        #    it may or may not have an integer.
        #    The 'locus_tag' qualifier may or may not be present,
        #    and may or may not have an integer.
        if value is None:
            value = ""
            list1 = ["_locus_tag_num", "_gene_num", "_product_num",
                     "_note_num", "_function_num"]

            # First see if any num attributes have a float.
            x = 0
            while (value == "" and x < len(list1)):
                name = getattr(self, list1[x])
                if basic.is_float(name):
                    value = name
                x += 1

            # Second see if any num attributes have non-empty values.
            # At this point, it's very unpredictable. Values could be like
            # 'terL' in the gene or like '10a' in the locus_tag.
            if value == "":
                list2 = ["gene"]
                list2.extend(list1)
                y = 0
                while (value == "" and y < len(list2)):
                    name = getattr(self, list2[y])
                    if name != "":
                        value = name
                    y += 1
        self.name = value


    def set_gene(self, value, delimiter=None, prefix_set=None):
        """Set the gene attribute.

        :param value: Gene data to parse. Also passed to set_num().
        :type value: str
        :param delimiter: Passed to set_num().
        :type delimiter: str
        :param prefix_set: Passed to set_num().
        :type prefix_set: set
        """
        value = value.strip()
        self.gene = value
        self.set_num("_gene_num", value,
                     delimiter=delimiter, prefix_set=prefix_set)


    def set_description_field(self, attr, description,
                              delimiter=None, prefix_set=None):
        """Set a description attribute parsed from a description.

        :param attr: Attribute to set the description.
        :type attr: str
        :param description: Description data to parse. Also passed to set_num().
        :type description: str
        :param delimiter: Passed to set_num().
        :type delimiter: str
        :param prefix_set: Passed to set_num().
        :type prefix_set: set
        """
        # Used to set product/raw_product, function/raw_function, note/raw_note.
        # attr  # e.g. self.product
        raw_attr = "raw_" + attr  # e.g. self.raw_product
        num_attr = "_" + attr + "_num"  # e.g. self._product_num
        raw, processed = basic.reformat_description(description)
        setattr(self, attr, processed)
        setattr(self, raw_attr, raw)
        self.set_num(num_attr, description,
                     delimiter=delimiter, prefix_set=prefix_set)


    def set_num(self, attr, description, delimiter=None, prefix_set=None):
        """Set a number attribute from a description.

        :param attr: Attribute to set the number.
        :type attr: str
        :param description: Description data from which to parse the number.
        :type description: str
        :param delimiter: Value used to split the description data.
        :type delimiter: str
        :param prefix_set: Valid possible delimiters in the description.
        :type prefix_set: set
        """
        # Used to set product_num, function_num, note_num, gene_num
        # Sometimes feature number is stored within the description
        # (e.g. 'gp10; terminase' or 'terminase; gp10')
        # List of delimiters found in description fields
        # ";" is probably the most common.
        delimiters = [";", ","]
        if delimiter is None:
            delimiter = basic.choose_most_common(description, delimiters)
        if prefix_set is None:
            prefix_set = {"gp", "orf", ""}

        # Iterate through the list of strings, and select the first
        # string that looks like a gene number.
        split_value = description.split(delimiter)
        num = ""
        x = 0
        while (num == "" and x < len(split_value)):
            string = split_value[x]
            string = string.strip()
            left, right = basic.split_string(string)
            # Possible returns:
            # 1. left is alpha and right is number or float (gp10 = gp, 10),
            # 2. string is alpha (terminase = terminase, ""),
            # 3. string is number (10 = "", 10).
            if left.lower() in prefix_set:
                num = right
            x += 1
        setattr(self, attr, num)


    def set_description(self, value):
        """Set the primary raw and processed description attributes.

        :param value:
            Indicates which reference attributes are used
            to set the attributes ('product', 'function', 'note').
        :type value: str
        """

        if value == "product":
            self.raw_description = self.raw_product
            self.description = self.product
        elif value == "function":
            self.raw_description = self.raw_function
            self.description = self.function
        elif value == "note":
            self.raw_description = self.raw_note
            self.description = self.note
        else:
            pass


    def get_begin_end(self):
        """Get feature coordinates in transcription begin-end format.

        :returns:
            (Begin, End) Start and stop coordinates ordered by which coordinate
            indicates the transcriptional beginning and end of the feature.
        :rtype: tuple
        """

        # Ensure format of orientation info.
        orientation = basic.reformat_strand(self.orientation, "fr_long")

        if orientation == "forward":
            begin = self.start
            end = self.stop
        elif orientation == "reverse":
            begin = self.stop
            end = self.start
        else:
            begin = -1
            end = -1

        return (begin, end)

    def translate_seq(self):
        """Translate the CDS nucleotide sequence.

        Use Biopython to translate the nucleotide sequece.
        The method expects the nucleotide sequence to be a valid CDS
        sequence in which:

            1. it begins with a valid start codon,
            2. it ends with a stop codon,
            3. it contains only one stop codon,
            4. its length is divisible by 3,
            5. it translates non-standard start codons to methionine.

        If these criteria are not met, an empty Seq object is returned.

        :returns: Amino acid sequence
        :rtype: Seq
        """

        try:
            translation = self.seq.translate(table=self.translation_table,
                                             cds=True)
        except:
            translation = Seq("", IUPAC.protein)
        return translation


    def set_translation(self, value=None, translate=False):
        """Set translation and its length.

        The translation is coerced into a Biopython Seq object.
        If no input translation value is provided, the translation is
        generated from the parent genome nucleotide sequence.
        If an input translation value is provided, the 'translate'
        parameter has no impact.

        :param value: Amino acid sequence
        :type value: str or Seq
        :param translate:
            Indicates whether the translation should be generated
            from the parent genome nucleotide sequence.
        :type translate: bool
        """

        if isinstance(value, Seq):
            self.translation = value.upper()
        elif value is not None:
            try:
                self.translation = Seq(value.upper(), IUPAC.protein)
            except:
                self.translation = Seq("", IUPAC.protein)
        elif translate:
            self.translation = self.translate_seq()
        else:
            self.translation = Seq("", IUPAC.protein)
        self.translation_length = len(self.translation)


    def set_translation_table(self, value):
        """Set translation table integer.

        :param value:
            Translation table that should be used to generate the translation.
        :type value: int
        """

        try:
            self.translation_table = int(value)
        except:
            self.translation_table = -1


    def set_orientation(self, value, format, case=False):
        """Sets orientation based on indicated format.

        Relies on the  `reformat_strand` function to manage orientation data.

        :param value: Input orientation value.
        :type value: misc.
        :param format: Indicates how the orientation data should be formatted.
        :type format: str
        :param case: Indicates whether the output orientation data should be cased.
        :type case: bool
        """
        self.orientation = basic.reformat_strand(value, format, case)


    def set_location_id(self):
        """Create a tuple of feature location data.

        For start and stop coordinates of the feature, it doesn't matter
        whether the feature is complex with a translational frameshift or not.
        Retrieving the "start" and "stop" boundary attributes return the very
        beginning and end of the feature, disregarding the
        inner "join" coordinates.
        If only the feature transcription "end" coordinate is used,
        orientation information is required.
        If transcription "begin" and "end" coordinates are used
        instead of "start" and "stop" coordinates,
        no orientation information is required.
        """
        self._start_stop_orient_id = (self.start, self.stop, self.orientation)

        begin, end = self.get_begin_end()
        self._end_orient_id = (end, self.orientation)
        self._start_end_id = (begin, end)


    def reformat_start_and_stop(self, new_format):
        """Convert start and stop coordinates to new coordinate format.
        This also updates the coordinate format attribute to reflect
        change.

        Relies on the `reformat_coordinates` function.

        :param new_format: Indicates how coordinates should be formatted.
        :type new_format: str
        """
        new_start, new_stop = basic.reformat_coordinates(
            self.start, self.stop, self.coordinate_format, new_format)
        if (new_format != self.coordinate_format):
            self.start = new_start
            self.stop = new_stop
            self.coordinate_format = new_format


    def set_nucleotide_length(self, seq=False, translation=False):
        """Set the length of the nucleotide sequence.

        Nucleotide length can be computed several different ways, including
        from the difference of the start and stop coordinates, the length
        of the transcribed nucleotide sequence, or the length of
        the translation. For compound features, using either the nucleotide or
        translation sequence is the accurate way to determine the
        true length of the feature, but 'length' may mean different things
        in different contexts.

        :param seq:
            Use the nucleotide sequence from the 'seq' attribute to
            compute the length.
        :type seq: bool
        :param translation:
            Use the translation sequence from the 'translation' attribute to
            compute the length.
        :type translation: bool
        """

        if seq:
            self.length = len(self.seq)
        elif translation:
            self.length = (len(self.translation) * 3) + 3 # include stop codon
        else:
            if self.coordinate_format == "0_half_open":
                self.length = self.stop - self.start
            elif self.coordinate_format == "1_closed":
                self.length = self.stop - self.start + 1
            else:
                self.length = -1


    def set_nucleotide_sequence(self, value=None, parent_genome_seq=None):
        """Set the nucleotide sequence of the feature.

        This method can directly set the attribute from a supplied 'value',
        or it can retrieve the sequence from the parent genome using
        Biopython. In this latter case, it relies on a Biopython SeqFeature
        object for the sequence extraction method and coordinates.
        If this object was generated from a Biopython-parsed
        GenBank-formatted flat file, the coordinates are by default
        '0-based half-open', the object contains coordinates for every
        part of the feature (e.g. if it is a compound feature) and
        fuzzy locations. As a result, the length of the retrieved sequence
        may not exactly match the length indicated from the 'start' and 'stop'
        coordinates.
        If the nucleotide sequence 'value' is provided, the
        'parent_genome_seq' does not impact the result.

        :param value: Input nucleotide sequence
        :type value: str of Seq
        :param parent_genome_seq: Input parent genome nucleotide sequence.
        :type parent_genome_seq: Seq
        """

        if isinstance(value, Seq):
            self.seq = value.upper()
        elif value is not None:
            try:
                self.seq = Seq(value.upper(), IUPAC.ambiguous_dna)
            except:
                self.seq = Seq("", IUPAC.ambiguous_dna)
        elif parent_genome_seq is not None:
            try:
                self.seq = self.seqfeature.extract(parent_genome_seq)
            except:
                # TODO not sure if this is the proper alphabet to use.
                self.seq = Seq("", IUPAC.ambiguous_dna)
        else:
            self.seq = Seq("", IUPAC.ambiguous_dna)



#Needs unittests, however:
#Seqfeature retrieval and generation is clunky probably requires some
#over arching seqfeature generation.
#This cannot be in flat_files due to the need to import Cds which would
#create circular dependancies.
#May delay unittests until structure is revamped
    def set_seqfeature(self, type="CDS"):
        """Set the 'seqfeature' attribute.

        The 'seqfeature' attribute stores a Biopython SeqFeature object,
        which contains methods valuable to extracting sequence data
        relevant to the feature.
        """
        # SeqFeature methods rely on coordinates in 0-based half-open
        # format and orientation to be numeric.
        new_start, new_stop = \
            basic.reformat_coordinates(self.start,
                                       self.stop,
                                       self.coordinate_format,
                                       "0_half_open")

        strand = basic.reformat_strand(self.orientation, "numeric")

        self.seqfeature = self.create_seqfeature(type, new_start, new_stop,
                                                                  strand)

    def create_seqfeature(self, type, start, stop, strand):
        if start <= stop:
            seqfeature = SeqFeature(FeatureLocation(start, stop),
                                            strand=strand, type=type)
        else:
            seqfeature = SeqFeature(CompoundLocation
                                  ([FeatureLocation(start, self.genome_length),
                                    FeatureLocation(0, stop)]),
                                            strand=strand, type=type)

        seqfeature.qualifiers = self.get_qualifiers(type)

        return seqfeature

    def get_qualifiers(self, type):
        """Helper function that uses cds data to populate
        the qualifiers SeqFeature attribute

        :returns:
            qualifiers(dictionary) is a dictionary with the
            formating of BioPython's SeqFeature qualifiers
            attribute.
        """
        qualifiers = OrderedDict()
        if self.description == "":
            product = "hypothetical protein"
        else:
            product = self.description

        if type == "CDS":
            qualifiers["gene"] = [self.name]
            if self.locus_tag != "":
                qualifiers["locus_tag"] = [self.locus_tag]
            qualifiers["note"] = ["".join(["gp", self.name])]
            qualifiers["codon_start"] = ["1"]
            qualifiers["transl_table"] = [self.translation_table]
            qualifiers["product"] = [product]
            qualifiers["translation"] = [self.translation]


        elif type == "Protein":
            qualifiers["product"] = [product]
        elif type == "gene":
           qualifiers["gene"] = [self.name]
           qualifiers["locus_tag"] = [self.locus_tag]

        return qualifiers




    # Evaluations.
    def set_eval(self, eval_id, definition, result, status):
        """Constructs and adds an Evaluation object to the evaluations list.

        :param eval_id: Unique identifier for the evaluation.
        :type eval_id: str
        :param definition: Description of the evaluation.
        :type definition: str
        :param result: Description of the outcome of the evaluation.
        :type result: str
        :param status: Outcome of the evaluation.
        :type status: str
        """
        evl = evaluation.Evaluation(eval_id, definition, result, status)
        self.evaluations.append(evl)


    def check_attribute(self, attribute, check_set, expect=False, eval_id=None,
                        success="correct", fail="error", eval_def=None):
        """Check that the attribute value is valid.

        :param attribute: Name of the CDS object attribute to evaluate.
        :type attribute: str
        :param check_set:
            Set of reference ids.
        :type check_set: set
        :param expect:
            Indicates whether the attribute value is expected to be present
            in the check set.
        :type expect: bool
        :param eval_id: Unique identifier for the evaluation.
        :type eval_id: str
        :param success: Default status if the outcome is a success.
        :type success: str
        :param fail: Default status if the outcome is not a success.
        :type fail: str
        :param eval_def: Description of the evaluation.
        :type eval_def: str
        """
        try:
            test = True
            value1 = getattr(self, attribute)
        except:
            test = False
            value1 = None
        if test:
            value1_short = basic.truncate_value(str(value1), 30, "...")
            result = f"The {attribute} value '{value1_short}' is "

            value2 = basic.check_value_expected_in_set(
                        value1, check_set, expect)
            if value2:
                result = result + "valid."
                status = success
            else:
                result = result + "not valid."
                status = fail
        else:
            result = f"'{attribute}' is not a valid attribute to be evaluated."
            status = "untested"
        definition = f"Check the value of the '{attribute}' attribute for {self.locus_tag} ({self.id})."
        definition = basic.join_strings([definition, eval_def])
        self.set_eval(eval_id, definition, result, status)


    def check_magnitude(self, attribute, expect, ref_value, eval_id=None,
                        success="correct", fail="error", eval_def=None):
        """Check that the magnitude of a numerical attribute is valid.

        :param attribute: same as for check_attribute().
        :param expect:
            Comparison symbol indicating direction of magnitude (>, =, <).
        :type expect: str
        :param ref_value: Numerical value for comparison.
        :type ref_value: int, float, datetime
        :param eval_id: same as for check_attribute().
        :param success: same as for check_attribute().
        :param fail: same as for check_attribute().
        :param eval_def: same as for check_attribute().
        """
        try:
            test = True
            query_value = getattr(self, attribute)
        except:
            test = False
            query_value = None
        if test:
            result = f"The {attribute} value {query_value} is "
            if query_value > ref_value:
                compare = ">"
                result = result + "greater than "
            elif query_value == ref_value:
                compare = "="
                result = result + "equal to "
            else:
                compare = "<"
                result = result + "less than "
            result = result + f"{ref_value}, which is "
            if compare == expect:
                result = result + "expected."
                status = success
            else:
                result = result + "not expected."
                status = fail
        else:
            result = f"'{attribute}' is not a valid attribute to be evaluated."
            status = "untested"
        definition = f"Check the magnitude of the '{attribute}' attribute for {self.locus_tag} ({self.id})."
        definition = basic.join_strings([definition, eval_def])
        self.set_eval(eval_id, definition, result, status)


    def check_translation(self, eval_id=None, success="correct",
                          fail="error", eval_def=None):
        """Check that the current and expected translations match.

        :param eval_id: same as for check_attribute().
        :param success: same as for check_attribute().
        :param fail: same as for check_attribute().
        :param eval_def: same as for check_attribute().
        """

        translation = self.translate_seq()
        exp_len = len(translation)
        result = f"The translation length ({self.translation_length}) "
        if self.translation_length < exp_len:
            result = result + f"is shorter than expected ({exp_len})."
            status = fail
        elif self.translation_length > exp_len:
            result = result + f"is longer than expected ({exp_len})."
            status = fail
        elif self.translation != translation:
            result = result + (f"is as expected, but the "
                               f"translation itself ({self.translation}) is "
                               f"different than expected ({translation}).")
            status = fail
        else:
            result = result + "and sequence are correct."
            status = success
        definition = f"Check that {self.locus_tag} ({self.id}) contains the expected translation."
        definition = basic.join_strings([definition, eval_def])
        self.set_eval(eval_id, definition, result, status)


    def check_amino_acids(self, check_set=set(), eval_id=None,
                          success="correct", fail="error", eval_def=None):
        """Check whether all amino acids in the translation are valid.

        :param check_set: Set of valid amino acids.
        :type check_set: set
        :param eval_id: same as for check_attribute().
        :param success: same as for check_attribute().
        :param fail: same as for check_attribute().
        :param eval_def: same as for check_attribute().
        """
        amino_acid_set = set(self.translation)
        amino_acid_error_set = amino_acid_set - check_set
        result = "The translation contains "
        if len(amino_acid_error_set) > 0:
            aae_string = ", ".join(amino_acid_error_set)
            result = result + f"the following unexpected amino acids: {aae_string}."
            status = fail
        else:
            result = result + "no unexpected amino acids."
            status = success

        definition = f"Check if all amino acids in the translation are expected for {self.locus_tag} ({self.id})."
        definition = basic.join_strings([definition, eval_def])
        self.set_eval(eval_id, definition, result, status)


    def check_orientation(self, format="fr_short", case=True, eval_id=None,
                          success="correct", fail="error", eval_def=None):
        """Check if orientation is set appropriately.

        Relies on the `reformat_strand` function to manage orientation data.

        :param format: Indicates how coordinates should be formatted.
        :type format: str
        :param case: Indicates whether the orientation data should be cased.
        :type case: bool
        :param eval_id: same as for check_attribute().
        :param success: same as for check_attribute().
        :param fail: same as for check_attribute().
        :param eval_def: same as for check_attribute().
        """
        expected_orient = basic.reformat_strand(self.orientation,
                                                format=format,
                                                case=case)
        result = f"The orientation is {self.orientation}, and it is formatted "
        if self.orientation == expected_orient:
            result = result + "correctly."
            status = success
        else:
            result = result + "incorrectly."
            status = fail
        definition = f"Check if the orientation is set correctly for {self.locus_tag} ({self.id})."
        definition = basic.join_strings([definition, eval_def])
        self.set_eval(eval_id, definition, result, status)


    def check_locus_tag_structure(self, check_value=None, only_typo=False,
                                  prefix_set=set(), case=True, eval_id=None,
                                  success="correct", fail="error", eval_def=None):
        """Check if the locus_tag is structured correctly.

        :param check_value:
            Indicates the genome id that is expected to be present.
            If None, the 'genome_id' parameter is used.
        :type check_value: str
        :param only_typo:
            Indicates if only the genome id spelling should be evaluated.
        :type only_typo: bool
        :param prefix_set:
            Indicates valid common prefixes, if a prefix is expected.
        :type prefix_set: set
        :param case:
            Indicates whether the locus_tag is expected to be capitalized.
        :type case: bool
        :param eval_id: same as for check_attribute().
        :param success: same as for check_attribute().
        :param fail: same as for check_attribute().
        :param eval_def: same as for check_attribute().
        """
        if check_value is None:
            check_value = self.genome_id

        results = []
        if only_typo:
            pattern = re.compile(check_value.lower())
            search_result = pattern.search(self.locus_tag.lower())
            if search_result == None:
                results.append("The genome ID is missing.")
        else:
            # Expected structure: SEA_TRIXIE_20
            parts = self.locus_tag.split("_")
            if case:
                if self.locus_tag != self.locus_tag.upper():
                    results.append("The capitalization is incorrect.")
            if len(parts) == 3:
                if prefix_set is not None:
                    if parts[0].upper() not in prefix_set:
                        results.append("The prefix is missing.")
                if parts[1].upper() != check_value.upper():
                    results.append("The genome ID is missing.")
                if not parts[2].isdigit():
                    results.append("The feature number is missing.")
            else:
                results.append("The locus_tag does not contain three parts.")
        result = f"The locus_tag qualifier is {self.locus_tag}. It is "
        if len(results) == 0:
            result = result + "structured correctly."
            status = success
        else:
            result = result + "not structured correctly. " + " ".join(results)
            status = fail
        definition = f"Check if the locus_tag qualifier is structured correctly for {self.locus_tag} ({self.id})."
        definition = basic.join_strings([definition, eval_def])
        self.set_eval(eval_id, definition, result, status)


    def check_gene_structure(self, eval_id=None, success="correct",
                             fail="error", eval_def=None):
        """Check if the gene qualifier contains an integer.

        :param eval_id: same as for check_attribute().
        :param success: same as for check_attribute().
        :param fail: same as for check_attribute().
        :param eval_def: same as for check_attribute().
        """
        try:
            value = int(self.gene)
        except:
            value = self.gene

        result = f"The gene qualifier is {self.gene}. It "
        if isinstance(value, int):
            result = result + "contains an integer, as expected."
            status = success
        else:
            result = result + "does not contain an integer, which is not expected."
            status = fail
        definition = f"Check if the gene qualifier contains an integer for {self.locus_tag} ({self.id})."
        definition = basic.join_strings([definition, eval_def])
        self.set_eval(eval_id, definition, result, status)


    def check_compatible_gene_and_locus_tag(self, eval_id=None,
                                            success="correct", fail="error",
                                            eval_def=None):
        """Check if gene and locus_tag attributes contain identical numbers.

        :param eval_id: same as for check_attribute().
        :param success: same as for check_attribute().
        :param fail: same as for check_attribute().
        :param eval_def: same as for check_attribute().
        """
        result = (f"The numbers in the gene ({self.gene}) and "
                  f"locus_tag ({self._locus_tag_num}) qualifiers are ")
        if self.gene == self._locus_tag_num:
            result = result + "consistent."
            status = success
        else:
            result = result + "not consistent."
            status = fail
        definition = f"Check if the gene and locus_tag numbers are consistent for {self.locus_tag} ({self.id})."
        definition = basic.join_strings([definition, eval_def])
        self.set_eval(eval_id, definition, result, status)


    def check_description_field(self, attribute="product", eval_id=None,
                                success="correct", fail="error", eval_def=None):
        """Check if there are CDS descriptions in unexpected fields.

        Evaluates whether the indicated attribute is empty or generic,
        and other fields contain non-generic data.

        :param attribute:
            Indicates the reference attribute for the evaluation
            ('product', 'function', 'note').
        :type attribute: str
        :param eval_id: same as for check_attribute().
        :param success: same as for check_attribute().
        :param fail: same as for check_attribute().
        :param eval_def: same as for check_attribute().
        """

        description = ""
        description_set = set()

        if attribute == "product":
            description = self.product
        else:
            if self.product != "":
                description_set.add(self.product)

        if attribute == "function":
            description = self.function
        else:
            if self.function != "":
                description_set.add(self.function)

        if attribute == "note":
            description = self.note
        else:
            if self.note != "":
                description_set.add(self.note)

        result = f"The CDS description is '{description}', "
        if (description == "" and len(description_set) > 0):
            d_string = ", ".join(description_set)
            result = result + ("but there is a non-generic description in "
                               "at least one other qualifier: "
                               f"'{d_string}'. This is not expected.")
            status = fail
        else:
            result = result + "as expected."
            status = success
        definition = ("Check if there is a discrepancy "
                      f"between description fields for {self.locus_tag} ({self.id}).")
        definition = basic.join_strings([definition, eval_def])
        self.set_eval(eval_id, definition, result, status)

    # TODO this should probably be implemented at the genome level,
    # calculating how many CDS features contain non-generic data.
    def check_generic_data(self, attribute=None, eval_id=None,
                           success="correct", fail="error", eval_def=None):
        """Check if the indicated attribute contains generic data.

        :param attribute:
            Indicates the attribute for the evaluation
            ('product', 'function', 'note').
        :type attribute: str
        :param eval_id: same as for check_attribute().
        :param success: same as for check_attribute().
        :param fail: same as for check_attribute().
        :param eval_def: same as for check_attribute().
        """

        if attribute == "product":
            original = self.raw_product
            processed = self.product
        elif attribute == "function":
            original = self.raw_function
            processed = self.function
        elif attribute == "note":
            original = self.raw_note
            processed = self.note
        else:
            original = ""
            processed = ""

        if original == processed:
            result = f"The '{attribute}' field is correct."
            status = success
        else:
            result = f"The '{attribute}' field is not correct."
            status = fail
        definition = f"Check if the '{attribute}' field contains generic data for {self.locus_tag} ({self.id})."
        definition = basic.join_strings([definition, eval_def])
        self.set_eval(eval_id, definition, result, status)



    # TODO can be implemented once there are pre-defined lists available
    # of acceptable and non-acceptable gene descriptions.
    # TODO unittest.
    # def check_valid_description(self, check_set=None, attribute=None,
    #                             eval_id=None):
    #     """Check if the CDS description in the indicated attribute is valid."""
    #     pass
